# coding=utf-8
import datetime

import django
from django.conf import settings
from ipywidgets import *
import mammut.common.universal_dependencies as ud
import os
from mammut.linguistics.concordance import Concordance
from functools import reduce
from collections import namedtuple
from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from mammut.common.lexicon.dictionary import (
    DictionaryDefinitionEntry,
    DictionaryDefinitionSetEntry,
    DictionaryPosTagEntry,
    DictionaryLemmaEntry
)

MODULE_PATH = os.path.realpath(os.path.dirname(__file__))

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [MODULE_PATH + """/templates"""],
    }
]
settings.configure(TEMPLATES=TEMPLATES, DEBUG=True)
django.setup()

from django.template import Template, Context
from django.template.loader import get_template
from .widgets_utilities import show_paradigm_application

pending_words_template = get_template("pending_words.html")
concordance_template = get_template("concordance.html")
external_dict_links_template = Template(
    """<html>
<body>
{% for edl in ext_dict_links %}
    <p><a href="{{ edl.link }}" target="_blank">{{ edl.id }}: {{ lemma }}</a></p>
{% endfor %}
</body>
</html>"""
)

PendingWordDesc = namedtuple("PendingWordDesc", ["text", "strike"])
ConcordanceDesc = namedtuple("ConcordanceDesc", ["morpheme", "text", "source"])
ParadigmAppWordDesc = namedtuple("ParadigmAppWordDesc", ["morpheme", "features"])
ExternalDictionaryLinksDesc = namedtuple("ExternalDictionaryLinksDesc", ["id", "link"])

DEFAULT_STANDARD = None


def create_dictionary_definition_section(
    standard: LinguisticStandard,
    def_entry: DictionaryDefinitionEntry,
    delete_callback,
    regional_settings,
):
    out_vocabulary_ratio = widgets.ToggleButton(
        value=True,
        description="OVR: 0/0",
        disabled=True,
        button_style="success",
        tooltip="Out of vocabulary ratio",
        icon="check",
        layout=Layout(width="100px"),
    )
    primes_ratio = widgets.ToggleButton(
        value=True,
        description="SPR: 0/0",
        disabled=True,
        button_style="success",
        tooltip="Semantic Primes Exponent Ratio",
        icon="check",
        layout=Layout(width="100px"),
    )
    in_vocabulary = widgets.VBox([out_vocabulary_ratio, primes_ratio])
    # entry object where to store the values selected
    if def_entry is None:
        def_entry = DictionaryDefinitionEntry(
            "",
            "",
            "",
            standard.regional_settings_definition_list[0],
            datetime.date.today().strftime("%Y/%m/%d"),
        )
        # TODO: hay que cambar esto para que use common.LOCAL_DATE_FORMATTER. Pero hacer esto implica ir a todos los diccionarios ya existentes y actualizarlos.
    # definition widget to show to the user
    definition = widgets.Textarea(
        value=def_entry.definition,
        placeholder="Definición",
        description="",
        layout=Layout(width="350px", height="70px"),
    )
    # details about the out of vocabulary words
    out_vocabulary_details = widgets.Textarea(
        value="",
        placeholder="",
        description="",
        disabled=True,
        layout=Layout(width="270px", height="35px"),
    )

    def change_vocabulary_info():
        inv = standard.is_in_vocabulary(def_entry.definition, regional_settings)
        if inv.value:
            out_vocabulary_ratio.description = inv.description
            out_vocabulary_ratio.value = True
            out_vocabulary_ratio.button_style = "success"
        else:
            out_vocabulary_ratio.description = inv.description
            out_vocabulary_ratio.value = False
            out_vocabulary_ratio.button_style = "danger"
        if inv.base_value:
            primes_ratio.description = inv.base_description
            primes_ratio.value = True
            primes_ratio.button_style = "success"
        else:
            primes_ratio.description = inv.base_description
            primes_ratio.value = False
            primes_ratio.button_style = "danger"
        out_vocabulary_details.value = inv.details

    def on_definition_value_change(change):
        def_entry.definition = change["new"]
        change_vocabulary_info()

    change_vocabulary_info()
    definition.observe(on_definition_value_change, names="value")
    # source widget to show to the user
    ext_dic_list = [
        ed.source_id for ed in standard.external_dictionary_sources.values()
    ]
    if def_entry.source == "":
        source_value = ext_dic_list[0]
    else:
        source_value = def_entry.source

    source = widgets.Dropdown(
        options=ext_dic_list,
        value=source_value,
        description="",
        layout=Layout(width="110px"),
    )
    def_entry.source = ext_dic_list[0]

    def on_source_value_change(change):
        def_entry.source = change["new"]

    source.observe(on_source_value_change, names="value")
    # topic widget to show to the user
    topic = widgets.Text(
        value=def_entry.topic,
        placeholder="",
        description="",
        layout=Layout(width="80px"),
    )

    def on_topic_value_change(change):
        def_entry.topic = change["new"]

    topic.observe(on_topic_value_change, names="value")
    # regional settings definition widget to show to the user
    if def_entry.regional_settings_definition == "":
        regional_settings_definition_value = standard.regional_settings_definition_list[
            0
        ]
    else:
        regional_settings_definition_value = def_entry.regional_settings_definition

    regional_settings_definition = widgets.Dropdown(
        options=standard.regional_settings_definition_list,
        value=regional_settings_definition_value,
        description="",
        layout=Layout(width="60px"),
    )

    def on_regional_settings_definition_value_change(change):
        def_entry.regional_settings_definition = change["new"]

    regional_settings_definition.observe(
        on_regional_settings_definition_value_change, names="value"
    )
    # delete button widget to show to the user
    delete = widgets.Button(
        description="-",
        button_style="",
        tooltip="Eliminar",
        layout=Layout(width="20px"),
    )
    pre_box1 = widgets.HBox([source, topic, regional_settings_definition, delete])
    pre_box2 = widgets.VBox([pre_box1, out_vocabulary_details])
    # the box control that group all the control that form the definition
    box = widgets.HBox([in_vocabulary, definition, pre_box2])

    def delete_on_click(sender):
        delete_callback(box, def_entry)

    delete.on_click(delete_on_click)
    return [box, def_entry]


def create_dictionary_definition_set_section(
    postag_value,
    definition_index: int,
    standard: LinguisticStandard,
    delete_callback,
    definition_set_entry_param: DictionaryDefinitionSetEntry,
    dictionary,
    lemma,
    regional_settings,
):
    if definition_set_entry_param is None:
        definition_set_entry = DictionaryDefinitionSetEntry(definition_index, None)
    else:
        definition_set_entry = definition_set_entry_param

    def get_features_list():
        res = []
        for f in definition_set_entry.features_list:
            res.append(standard.feature_descriptor_list[f])
        return res

    features = widgets.SelectMultiple(
        options=standard.feature_descriptor_by_pos_tag[postag_value.id],
        value=([] if definition_set_entry_param is None else get_features_list()),
        description="Adicionales:",
        rows=3,
        layout=Layout(width="350px"),
    )

    def on_features_change(change):
        definition_set_entry.set_features_list(change["new"])

    features.observe(on_features_change, names="value")
    add_definition = widgets.Button(
        description="Agregar Parafrasis",
        button_style="",
        tooltip="Agregar Parafrasis",
        icon="check",
    )
    features_box = widgets.HBox([features, add_definition])
    dummy_label = widgets.Label(
        value="->", placeholder="", description="", layout=Layout(width="100px")
    )
    definition_label = widgets.Label(
        value="Definición",
        placeholder="Definición",
        description="Definición",
        layout=Layout(width="350px"),
    )
    source_label = widgets.Label(
        value="Fuente",
        placeholder="Fuente",
        description="Fuente",
        layout=Layout(width="110px"),
    )
    topic_label = widgets.Label(
        value="Topico",
        placeholder="Topico",
        description="Topico",
        layout=Layout(width="80px"),
    )
    regional_settings_def_label = widgets.Label(
        value="Idioma",
        placeholder="Idioma",
        description="Idioma",
        layout=Layout(width="60px"),
    )
    box_definition_label = widgets.HBox(
        [
            dummy_label,
            definition_label,
            source_label,
            topic_label,
            regional_settings_def_label,
        ]
    )
    definitions_box = widgets.VBox([box_definition_label])
    box = widgets.VBox([features_box, definitions_box])

    def delete_definition_callback(definition_box, def_entry):
        definitions_box.children = list(
            filter(lambda db: db != definition_box, definitions_box.children)
        )
        definition_set_entry.definitions_entries.remove(def_entry)
        dictionary.delete_definition_lemma(lemma, def_entry.definition)
        if len(definitions_box.children) == 1:
            delete_callback(box, definition_set_entry)

    def add_defition():
        definition_box, def_entry = create_dictionary_definition_section(
            standard, None, delete_definition_callback, regional_settings
        )
        definitions_box.children = [c for c in definitions_box.children] + [
            definition_box
        ]
        definition_set_entry.definitions_entries.append(def_entry)

    if definition_set_entry_param is None:
        add_defition()

    def add_definition_if_exist():
        if len(definition_set_entry.definitions_entries) > 0:
            for i in definition_set_entry.definitions_entries:
                definition_box, def_entry = create_dictionary_definition_section(
                    standard, i, delete_definition_callback, regional_settings
                )
                definitions_box.children = [c for c in definitions_box.children] + [
                    definition_box
                ]
                # definition_set_entry.definitions_entries.append(def_entry)

    if definition_set_entry_param is not None:
        add_definition_if_exist()

    def handle_add_definition(sender):
        add_defition()

    add_definition.on_click(handle_add_definition)
    return [box, definition_set_entry]


def create_dictionary_pos_tag_section(
    lemma: str,
    postag_value,
    postag_change_callback,
    delete_callback,
    standard: LinguisticStandard,
    index_name: str,
    pos_tag_entry_param: DictionaryPosTagEntry,
    regional_settings,
    dictionary,
):
    if pos_tag_entry_param is None:
        pos_tag_entry = DictionaryPosTagEntry(postag_value.id, None)
        morph_parad_value = ""
    else:
        pos_tag_entry = pos_tag_entry_param
        morph_parad_value = pos_tag_entry.morpheme_paradigm
    pos_tag = widgets.Dropdown(
        options=list(standard.pos_descriptor_list.values()),
        value=postag_value,
        description="PoS:",
        layout=Layout(width="260px"),
    )
    morph_parad_vals, morph_paradigms = standard.get_morpheme_paradigms(
        postag_value.id, regional_settings
    )
    if morph_parad_vals.__len__() == 0:
        options_value = {"": ""}
    else:
        options_value = {}
        for i, morph in enumerate(morph_paradigms):
            if morph_parad_value == "":
                morph_parad_value = morph.word_model
            morph_parad_key = (
                morph_parad_vals[i]
                + "        "
                + reduce(
                    lambda a, b: a + " | " + b
                    if b is not ""
                    else a + " | " + "sin descripcion",
                    list(morph.descriptions.values()),
                    "",
                )
            )
            options_value[morph_parad_key] = morph.word_model

    morph_parad = widgets.Dropdown(
        options=options_value,
        value=morph_parad_value,
        description="Parad. Morfema:",
        layout=Layout(width="320px"),
    )
    pos_tag_entry.morpheme_paradigm = morph_parad_value

    def morph_parad_change(change):
        pos_tag_entry.morpheme_paradigm = change["new"]

    morph_parad.observe(morph_parad_change, names="value")
    morph_parad_apply = widgets.Button(
        description="",
        button_style="",
        tooltip="Aplicar Paridigma de Morfema",
        icon="check",
        disabled=(len(morph_parad_vals) == 0),
        layout=Layout(width="20px"),
    )
    morph_parad_box = widgets.HBox([morph_parad, morph_parad_apply])
    miss_parad_vals = standard.get_misspelling_paradigms(
        postag_value.id, regional_settings
    )
    miss_parad = widgets.Dropdown(
        options=miss_parad_vals,
        description="Parad. Errores:",
        layout=Layout(width="320px"),
    )
    pos_tag_entry.misspellings_paradigm = ""

    def miss_parad_change(change):
        pos_tag_entry.misspellings_paradigm = change["new"]

    miss_parad.observe(miss_parad_change, names="value")
    miss_parad_apply = widgets.Button(
        description="",
        button_style="",
        tooltip="Aplicar Paridigma de Errores",
        icon="check",
        disabled=(len(miss_parad_vals) == 0),
        layout=Layout(width="20px"),
    )
    miss_parad_box = widgets.HBox([miss_parad, miss_parad_apply])
    fet_parad_vals = standard.get_feature_paradigms(postag_value.id, regional_settings)
    fet_parad = widgets.Dropdown(
        options=fet_parad_vals,
        description="Otros Parad.:",
        layout=Layout(width="320px"),
    )
    pos_tag_entry.features_paradigm = ""

    def fet_parad_change(change):
        pos_tag_entry.features_paradigm = change["new"]

    fet_parad.observe(fet_parad_change, names="value")
    fet_parad_apply = widgets.Button(
        description="",
        button_style="",
        tooltip="Aplicar Paridigma Adicionales",
        icon="check",
        disabled=(len(fet_parad_vals) == 0),
        layout=Layout(width="20px"),
    )
    fet_parad_box = widgets.HBox([fet_parad, fet_parad_apply])
    show_concordance = widgets.Button(
        description="Concordancia",
        button_style="",
        tooltip="Ver Concordancia",
        icon="check",
        layout=Layout(width="100px"),
    )
    add_definition_set = widgets.Button(
        description="Agregar Def.",
        button_style="",
        tooltip="Agregar Definicion",
        icon="check",
        layout=Layout(width="100px"),
    )
    delete = widgets.Button(
        description="Borrar",
        button_style="",
        tooltip="Borrar",
        icon="",
        layout=Layout(width="60px"),
    )
    show_add_delete_box = widgets.HBox([show_concordance, add_definition_set, delete])
    pos_tag_parads_box = widgets.VBox(
        [pos_tag, morph_parad_box, miss_parad_box, fet_parad_box, show_add_delete_box]
    )
    parad_application = widgets.HTML(value="")

    def morph_parad_apply_click(sender):
        morph_value = morph_parad.value.split(" ")[0]
        morpheme_res = standard.get_morphemes(
            pos_tag.value.id, morph_value, lemma, regional_settings
        )
        word_rows = [
            ParadigmAppWordDesc(
                m.text, ", ".join(map(lambda f: f.feature_name, m.features))
            )
            for m in morpheme_res
        ]
        show_paradigm_application(
            "Paradigma Morfema Aplicado", word_rows, parad_application
        )
        return word_rows

    morph_parad_apply.on_click(morph_parad_apply_click)

    def miss_parad_apply_click(sender):
        morpheme_res = standard.get_misspelling(
            pos_tag.value.id, miss_parad.value, lemma, regional_settings
        )
        word_rows = [
            ParadigmAppWordDesc(
                m.text, ", ".join(map(lambda f: f.feature_name, m.features))
            )
            for m in morpheme_res
        ]
        show_paradigm_application(
            "Paradigma Errores Aplicado", word_rows, parad_application
        )

    miss_parad_apply.on_click(miss_parad_apply_click)

    def fet_parad_apply_click(sender):
        morpheme_res = standard.get_other_inflections(
            pos_tag.value.id, fet_parad.value, lemma, regional_settings
        )
        word_rows = [
            ParadigmAppWordDesc(
                m.text, ", ".join(map(lambda f: f.feature_name, m.features))
            )
            for m in morpheme_res
        ]
        show_paradigm_application(
            "Paradigma Adicionales Aplicado", word_rows, parad_application
        )

    fet_parad_apply.on_click(fet_parad_apply_click)

    show_paradigm_application("Ningun paradigma aplicado", [], parad_application)
    pos_tag_parads_app_box = widgets.HBox([pos_tag_parads_box, parad_application])
    context_dict = {"concordance_rows": []}
    concordance_html = concordance_template.render(context_dict)
    concordance = widgets.HTML(value=concordance_html)
    definitions_box = widgets.VBox([])
    accordion = widgets.Accordion(children=[concordance, definitions_box])
    accordion.set_title(0, "Concordancia")
    accordion.set_title(1, "Definiciones")

    def show_concordance_click(sender):
        concordance_rows = [
            ConcordanceDesc("Implementar", "Falta implementar esto.", "Mammut"),
            ConcordanceDesc(
                "Implementare", "Lo implementare con un Web Scrapper.", "Mammut"
            ),
            ConcordanceDesc(
                "Implementado", "Tan pronto este implementado, lo subire.", "Mammut"
            ),
        ]
        context_dict = {"concordance_rows": concordance_rows}
        # TODO: revisar bien para eliminar data dummy
        concordance_results = []
        concordance_lib = Concordance()
        words_list = morph_parad_apply_click(
            sender
        )  # TODO:deberia ser con todos las aplicaiones de paradigmas, incluyendo errores y otros
        for i in words_list:
            word_result = concordance_lib.get_data_from_index(index_name, i.morpheme)
            if len(word_result):
                concordance_results.append(word_result)
        concordance_html = concordance_template.render(
            {"concordance_results": concordance_results}
        )
        concordance.value = concordance_html

    show_concordance.on_click(show_concordance_click)

    def delete_definition_callback(definition_set_box, def_set_entry):
        definitions_box.children = list(
            filter(lambda db: db != definition_set_box, definitions_box.children)
        )
        pos_tag_entry.definitions_set_entries.remove(def_set_entry)

    def add_definition_set_click(sender):
        definition_box, definition_set_entry = create_dictionary_definition_set_section(
            postag_value,
            len(definitions_box.children),
            standard,
            delete_definition_callback,
            None,
            dictionary,
            lemma,
            regional_settings,
        )
        definitions_box.children = [c for c in definitions_box.children] + [
            definition_box
        ]
        pos_tag_entry.definitions_set_entries.append(definition_set_entry)

    add_definition_set.on_click(add_definition_set_click)

    def add_definition_set_if_exist():
        if len(pos_tag_entry.definitions_set_entries) > 0:
            for i in pos_tag_entry.definitions_set_entries:
                (
                    definition_box,
                    definition_set_entry,
                ) = create_dictionary_definition_set_section(
                    postag_value,
                    len(definitions_box.children),
                    standard,
                    delete_definition_callback,
                    i,
                    dictionary,
                    lemma,
                    regional_settings,
                )

                definitions_box.children = [c for c in definitions_box.children] + [
                    definition_box
                ]
                # pos_tag_entry.definitions_set_entries.append(definition_set_entry)

    if pos_tag_entry_param is not None:
        add_definition_set_if_exist()

    def postag_change(change):
        pos_tag_entry.pos_tag = change["new"].id
        ##morph_parad_vals = list(standard.morpheme_paradigms[pos_tag.value.id][regional_settings].producers.keys())
        morph_parad_vals, morph_paradigms = standard.get_morpheme_paradigms(
            pos_tag.value.id, regional_settings
        )
        morph_parad_value = ""
        if morph_parad_vals.__len__() == 0:
            options_value = {"": ""}
        else:
            options_value = {}
            for i, morph in enumerate(morph_paradigms):
                if morph_parad_value == "":
                    morph_parad_value = morph.word_model
                morph_parad_key = (
                    morph_parad_vals[i]
                    + "        "
                    + reduce(
                        lambda a, b: a + " | " + b
                        if b is not ""
                        else a + " | " + "sin descripcion",
                        list(morph.descriptions.values()),
                        "",
                    )
                )
                options_value[morph_parad_key] = morph.word_model
        morph_parad.options = options_value
        morph_parad.value = morph_parad_value
        miss_parad_vals = list(
            standard.misspelling_paradigms[pos_tag.value.id][""].producers.keys()
        )
        miss_parad.options = miss_parad_vals
        fet_parad_vals = list(standard.feature_paradigms[pos_tag.value.id].keys())
        fet_parad.options = fet_parad_vals
        postag_change_callback(change)

    pos_tag.observe(postag_change, names="value")
    box = widgets.VBox([pos_tag_parads_app_box, accordion])

    def delete_click(sender):
        delete_callback(box, pos_tag_entry)

    delete.on_click(delete_click)

    return [box, pos_tag_entry]


def create_external_dictionaries_section(
    lemma: str, standard: LinguisticStandard, regional_settings
):
    ext_dict_html = []
    ext_dict_desc = []
    for ed in standard.external_dictionary_sources.values():
        if ed.source_url_pattern:
            if regional_settings in ed.regional_settings.split(","):
                ext_dict_html.append(widgets.HTML(value=ed.get_iframe(lemma)))
                ext_dict_desc.append(
                    ExternalDictionaryLinksDesc(ed.source_id, ed.get_url(lemma))
                )
    context_dict = {"lemma": lemma, "ext_dict_links": ext_dict_desc}
    ext_dict_html.append(
        widgets.HTML(value=external_dict_links_template.render(Context(context_dict)))
    )

    accordion = widgets.Accordion(children=ext_dict_html, tooltip="accordion")
    for i in range(0, len(ext_dict_desc)):
        accordion.set_title(i, ext_dict_desc[i].id)
    accordion.set_title(len(ext_dict_desc), "LINKS")
    return accordion


def create_lemma_section(dictionary, index_name: str):
    default_lemma = "mammut"
    lemma_entry = DictionaryLemmaEntry(default_lemma, None)
    # creamos el control donde se debe introducir el Lemma
    lemma_text = widgets.Text(
        value="",
        placeholder="Forma canonica de la palabra",
        description="Lemma:",
        layout=Layout(width="250px"),
    )

    def on_lemma_text_value_change(change):
        lemma_entry.lemma = change["new"].lower()

    lemma_text.observe(on_lemma_text_value_change, names="value")
    # creamos el control donde se debe introducir el AFI
    afi_text = widgets.Text(
        value="",
        placeholder="AFI de la forma canonica del lema",
        description="AFI:",
        layout=Layout(width="250px"),
    )

    def on_afi_text_value_change(change):
        lemma_entry.afi = change["new"].lower()

    afi_text.observe(on_afi_text_value_change, names="value")
    # creamos el dropdown para el tipo de palabra
    word_type = widgets.Dropdown(
        options=dictionary.standard.word_type_list,
        value=dictionary.standard.word_type_list[0],
        description="Tipo:",
        layout=Layout(width="220px"),
    )
    lemma_entry.word_type = dictionary.standard.word_type_list[0]

    def word_type_change(change):
        lemma_entry.word_type = change["new"]

    word_type.observe(word_type_change, names="value")
    # creamos el dropdown para el idioma del lemma
    reg_set_lemma = widgets.Dropdown(
        options=dictionary.standard.regional_settings_lemma_list,
        value=dictionary.standard.regional_settings_lemma_list[0],
        description="",
        layout=Layout(width="60px"),
        tooltip="regional_settings",
    )
    lemma_entry.regional_settings_lemma = dictionary.standard.regional_settings_lemma_list[
        0
    ]

    def reg_set_lemma_change(change):
        lemma_entry.regional_settings_lemma = change["new"]

    reg_set_lemma.observe(reg_set_lemma_change, names="value")
    # creamos la ayuda de los diccionarios en linea
    accordion = create_external_dictionaries_section(
        default_lemma, dictionary.standard, reg_set_lemma.value
    )
    search_lemma = widgets.Button(
        description="Buscar Lemma",
        button_style="",
        tooltip="Buscar Lemma",
        icon="",
        layout=Layout(width="100px"),
    )

    add_pos = widgets.Button(
        description="Agregar POS",
        button_style="",
        tooltip="Agregar POS",
        icon="",
        layout=Layout(width="100px"),
    )
    box_lemma_search_add = widgets.HBox(
        [lemma_text, afi_text, word_type, reg_set_lemma, search_lemma, add_pos]
    )
    # creamos el tab donde estaran los distintos morphemes
    pos_tab = widgets.Tab([])
    pos_values = list(dictionary.standard.pos_descriptor_list.values())
    pos_values_used = {}

    def on_pos_change_partial(index):
        def on_pos_change_complete(change):
            pos_tab.set_title(index, change["new"].name)
            pos_values.remove(change["new"])
            pos_values.append(pos_values_used[index])
            pos_values_used[index] = change["new"]

        return on_pos_change_complete

    def delete_pos_tab(pos_box, pos_tag_entry):
        pos_tab.children = list(filter(lambda pb: pb != pos_box, pos_tab.children))
        lemma_entry.remove_pos_tag_entry(pos_tag_entry)
        pos_values.append(
            dictionary.standard.pos_descriptor_list[pos_tag_entry.pos_tag]
        )
        add_pos.disabled = False

    def add_pos_tab(
        lemma,
        pos_tab_index,
        append_pos_entry,
        pos_tag_entry_param: DictionaryPosTagEntry,
    ):

        if pos_tag_entry_param is not None:
            in_postag_entry_param = True
            for index, i in enumerate(pos_values):
                if (
                    i.id == pos_tag_entry_param.pos_tag
                    and i not in pos_values_used.values()
                ):
                    pos_value = pos_values.pop(index)
                    pos_values_used[pos_tab_index] = pos_value
                    in_postag_entry_param = False
                    break
            if in_postag_entry_param:
                for index, i in enumerate(pos_values):
                    if i not in pos_values_used.values():
                        pos_value = pos_values.pop(index)
                        pos_values_used[pos_tab_index] = pos_value
                        break
        else:
            for index, i in enumerate(pos_values):
                if i not in pos_values_used.values():
                    pos_value = pos_values.pop(index)
                    pos_values_used[pos_tab_index] = pos_value
                    break

        if len(pos_values) == 0:
            add_pos.disabled = True
        pos_box, pos_tag_entry = create_dictionary_pos_tag_section(
            lemma,
            pos_value,  # TODO: en vez de pasar lemma, deberia pasar lemma_entry y listo, lemma es un atributo de ese objeto!
            on_pos_change_partial(pos_tab_index),
            delete_pos_tab,
            dictionary.standard,
            index_name,
            pos_tag_entry_param,
            reg_set_lemma.value,
            dictionary,
        )

        pos_tab.children = [c for c in pos_tab.children] + [pos_box]
        pos_tab.set_title(pos_tab_index, pos_value.name)
        if append_pos_entry:
            lemma_entry.add_pos_tag_entry(pos_tag_entry)
        lemma_text.disabled = True
        reg_set_lemma.disabled = True

    def handle_add_pos_tab(sender):
        # hacer busqueda en local y cambiar lemma entry por el lemm_entry si ya existia
        existing_lemma_entry = dictionary.get_lemma(lemma_entry.lemma)
        if existing_lemma_entry is not None:
            # set afi in input
            box_lemma_search_add.children[1].value = existing_lemma_entry.afi
            for i in lemma_entry.__dict__:
                lemma_entry.__dict__[i] = existing_lemma_entry.__dict__[i]

        if existing_lemma_entry is not None and len(pos_tab.children) == 0:
            box_buttons.children = [update_def, create_new_dic]
            for i in existing_lemma_entry.postags_entries:
                add_pos_tab(existing_lemma_entry.lemma, len(pos_tab.children), False, i)
        elif existing_lemma_entry is not None and len(pos_tab.children) > 0:
            box_buttons.children = [update_def, create_new_dic]
            add_pos_tab(existing_lemma_entry.lemma, len(pos_tab.children), True, None)
        else:
            box_buttons.children = [append_def, create_new_dic]
            add_pos_tab(lemma_entry.lemma, len(pos_tab.children), True, None)

    add_pos.on_click(handle_add_pos_tab)
    # creamos los botones de accion final
    box_buttons = widgets.HBox([])
    append_def = widgets.Button(
        description="Agregar Definiciones al final",
        button_style="",
        tooltip="Agregar Definiciones al final",
        icon="check",
        layout=Layout(width="300px"),
    )
    update_def = widgets.Button(
        description="Editar y Agregar Definiciones al final",
        button_style="",
        tooltip="Agregar Definiciones al final",
        icon="check",
        layout=Layout(width="300px"),
    )
    create_new_dic = widgets.Button(
        description="Crear Diccionario Nuevo",
        button_style="",
        tooltip="Crear Diccionario Nuevo",
        icon="check",
        layout=Layout(width="300px"),
    )
    all_box_childrens = [box_lemma_search_add, accordion, pos_tab, box_buttons]
    all_box = widgets.VBox(all_box_childrens)

    def handle_append_def(sender):
        for b in pos_tab.children:
            b.close()
        pos_tab.children = []
        box_buttons.children = []
        del pos_values[0:]
        for i in list(dictionary.standard.pos_descriptor_list.values()):
            pos_values.append(i)
        pos_values_used.clear()
        try:
            receipt = dictionary.add_lemma_entry(lemma_entry, True)
            print(receipt)
        except Exception as ex:
            print("el paradigma no es valido para el lemma, error:", ex)

    def handle_update_def(sender):
        for b in pos_tab.children:
            b.close()
        pos_tab.children = []
        box_buttons.children = []
        del pos_values[0:]
        for i in list(dictionary.standard.pos_descriptor_list.values()):
            pos_values.append(i)
        pos_values_used.clear()
        try:
            receipt = dictionary.update_lemma_entry(lemma_entry, True)
            print(receipt)
        except Exception as ex:
            print("el paradigma no es valido para el lemma, error:", ex)

    def handle_create_new_dic(sender):
        result = dictionary.create_new_dictionary(reg_set_lemma.value)
        print(f"Resultado de crear nuevo diccionario: {result}")

    append_def.on_click(handle_append_def)
    update_def.on_click(handle_update_def)
    create_new_dic.on_click(handle_create_new_dic)
    return all_box


def show_restart_add_lemma_entry_section_button(dictionary, index_name: str):
    # TODO:este boton es necesario porque algo pasa cuando se van a agregar un objeto nuevo no estan vaccios los
    # definitions_entries y se generan el numero de objetos igual a los que que se han creado desde que se ejecuto
    # la celda sin reiniciarla
    add_lemma_entry_section = create_lemma_section(dictionary, index_name)
    restart_lemma_section_button = widgets.Button(
        description="Reiniciar",
        button_style="",
        tooltip="Reiniciar",
        icon="check",
        layout=Layout(width="900px"),
    )

    all_box_childrens = [restart_lemma_section_button, add_lemma_entry_section]
    all_box = widgets.VBox(all_box_childrens)
    return all_box


# Utilitarian functions


# Universal dependecies


def create_universal_dependencies_help_section():
    def get_postag_help_links_html(pos_tag):
        return """<html>
    <body>
    <p><a href="{0}" target="_blank">PoS Tags - Ayuda General</a></p>

    <p><a href="{1}" target="_blank">PoS Tags - Estadisticas Para Español</a></p>
    </body>
    </html>""".format(
            ud.pos_tags_description_sites[pos_tag], ud.pos_tags_stats_sites_es[pos_tag]
        )

    pos_tag_desc_dd = widgets.Dropdown(
        options=ud.pos_tags,
        value=ud.noun_pos_tag,
        description="Tipo de Palabra:",
        layout=Layout(width="300px"),
    )
    pos_tag_help_links_html = widgets.HTML(
        value=get_postag_help_links_html(ud.noun_pos_tag)
    )
    pos_tag_desc_html = widgets.HTML(
        value="<iframe src={0} width=1000 height=350></iframe>".format(
            ud.pos_tags_description_sites[ud.noun_pos_tag]
        )
    )
    pos_tag_stat_html = widgets.HTML(
        value="<iframe src={0} width=1000 height=350></iframe>".format(
            ud.pos_tags_stats_sites_es[ud.noun_pos_tag]
        )
    )
    pos_tag_desc_box = widgets.VBox(
        [pos_tag_desc_dd, pos_tag_help_links_html, pos_tag_desc_html, pos_tag_stat_html]
    )

    def get_features_help_links_html(feature_selected):
        return """<html>
    <body>
    <p><a href="{0}" target="_blank">Features - Ayuda General</a></p>

    <p><a href="{1}" target="_blank">Features - Estadisticas Para Español</a></p>
    </body>
    </html>""".format(
            feature_selected.url_doc, feature_selected.url_stats_es
        )

    default_feature_selected = ud.pronominal_types_descriptor
    feature_descriptor_dd = widgets.Dropdown(
        options=ud.feature_types_descriptor_per_id,
        value=default_feature_selected,
        description="Tipo de Palabra:",
        layout=Layout(width="300px"),
    )
    feature_name = widgets.Text(
        value=default_feature_selected.name,
        description="Name:",
        disabled=True,
        layout=Layout(width="300px"),
    )
    feature_description = widgets.Text(
        value=default_feature_selected.description,
        description="Description:",
        disabled=True,
        layout=Layout(width="300px"),
    )
    box_feature_texts = widgets.HBox(
        [feature_descriptor_dd, feature_name, feature_description]
    )
    feature_help_links_html = widgets.HTML(
        value=get_features_help_links_html(default_feature_selected)
    )
    feature_descriptor_html = widgets.HTML(
        value="<iframe src={0} width=1000 height=350></iframe>".format(
            default_feature_selected.url_doc
        )
    )
    feature_stat_html = widgets.HTML(
        value="<iframe src={0} width=1000 height=350></iframe>".format(
            default_feature_selected.url_stats_es
        )
    )
    feature_descriptor_box = widgets.VBox(
        [
            box_feature_texts,
            feature_help_links_html,
            feature_descriptor_html,
            feature_stat_html,
        ]
    )

    def on_drop_down_change(change):
        if change["owner"] == pos_tag_desc_dd:
            pos_tag_help_links_html.value = get_postag_help_links_html(change["new"])
            pos_tag_desc_html.value = "<iframe src={0} width=1000 height=350></iframe>".format(
                ud.pos_tags_description_sites[change["new"]]
            )
            pos_tag_stat_html.value = "<iframe src={0} width=1000 height=350></iframe>".format(
                ud.pos_tags_stats_sites_es[change["new"]]
            )
        elif change["owner"] == feature_descriptor_dd:
            selected_feature = ud.feature_types_descriptor_per_id[change["new"][0]]
            feature_name.value = selected_feature.name
            feature_description.value = selected_feature.description
            feature_help_links_html.value = get_features_help_links_html(
                selected_feature
            )
            feature_descriptor_html.value = "<iframe src={0} width=1000 height=350></iframe>".format(
                selected_feature.url_doc
            )
            feature_stat_html.value = "<iframe src={0} width=1000 height=350></iframe>".format(
                selected_feature.url_stats_es
            )

    pos_tag_desc_dd.observe(on_drop_down_change, names="value")
    feature_descriptor_dd.observe(on_drop_down_change, names="value")

    accordion = widgets.Accordion(children=[pos_tag_desc_box, feature_descriptor_box])
    accordion.set_title(0, "Ayuda y Estadisticas de Part of Speechs")
    accordion.set_title(1, "Ayuda y Estadisticas de Features")
    return accordion



from datasets import load_dataset
from datasets import Dataset


class TextOnly:
    """TextOnly class is one of the dataset
    possibilities that can be used in mammut
    processes.
    """
    def data_translator(self):
        """Returns a pandas dataframe with the data of the
        single feature dataset.
        """
        sentences: Dataset = load_dataset("ptb_text_only", subset=None, split="train")
        sentences_df = sentences.data.to_pandas()
        return sentences_df



from mammut.common.simulator_utilities import *
from ipywidgets import *
from django.template.loader import get_template

from mammut.common.mammut_api import UserType
from mammut.common.package import Package, log
from mammut.common.simulator_utilities import to_sp_string


def get_compilation_boxes(
    simulator_obj, package_file, standard_spreadsheet_id, move_box_callback
):
    # Select your mammut layout #
    mammut_user_dropdown = widgets.Dropdown(
        description="Mammut bot", disabled=False, button_style=""
    )
    mammut_user_dropdown.options = simulator_obj.users[UserType.machine.value]
    select_mammut_user_btn = widgets.Button(
        description="Seleccionar", button_style="primary"
    )

    def select_user_button_clicked(b):
        simulator_obj.current_user = simulator_obj.mammut_api.get_user(
            mammut_user_dropdown.value
        )
        move_box_callback()

    select_mammut_user_btn.on_click(select_user_button_clicked)
    select_mammut_box = widgets.VBox(
        [widgets.HBox([mammut_user_dropdown, select_mammut_user_btn])]
    )
    # /Select your mammut layout #

    # Validate package #
    validate_package_html = widgets.HTML(value="")
    validate_package_file_box = widgets.Text(
        value=package_file, description="Id Archivo:", layout=Layout(width="100%")
    )
    validate_package_btn = widgets.Button(
        description="Validar Archivo", button_style="primary"
    )

    def validate_package_btn_clicked(b):
        try:
            # clear_output()
            # validar archivo
            validate_package_html.value = (
                "Validando archivo... <br> Por favor espere..."
            )
            simulator_obj.package = Package(
                validate_package_file_box.value,
                standard_spreadsheet_id
            )
            validate_package_html.value = "El archivo " + to_sp_string(True) + "<br>"
            move_box_callback()
        except Exception as ex:
            validate_package_html.value = "Error validando archivo... <br> " + str(ex)
            log.exception("Error validating file")

    update_simulator_corpus_list = widgets.Text(
        value="1",
        description="Corpus Id's (csv)",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    validate_package_btn.on_click(validate_package_btn_clicked)
    validate_package_box = widgets.VBox(
        [
            widgets.HBox(
                [
                    validate_package_html,
                    validate_package_file_box,
                    validate_package_btn,
                ],
                layout=Layout(width="100%", align_items="center"),
            )
        ]
    )
    # /Validate package #

    # /Update simulator #
    update_simulator_html = widgets.HTML(
        value='Presione "Actualizar Simulador" para guardar los cambios del excel en el servidor Mammut.'
    )
    update_simulator_html.layout = Layout(width="50%", height="100%")
    update_simulator_btn = widgets.Button(
        description="Actualizar Simulador", button_style="primary"
    )

    def get_update_button_clicked():
        def update_output_html_callback(mess, output):
            if output is not None:
                update_simulator_html.value += mess + " </br>" + output + " </br>"
            else:
                update_simulator_html.value += mess + " </br>"

        def update_button_clicked(b):
            update_simulator_html.value = (
                "Iniciando proceso de compilacion... </br> </br>"
            )
            corpus_list_to_compile = [
                int(x) for x in update_simulator_corpus_list.value.split(",")
            ]
            simulator_obj.package.compile_corpus_list(
                simulator_obj.current_user.id,
                update_output_html_callback,
                corpus_list_to_compile,
            )
            # TODO: ELiminar o cambiar por el nuevo compilador
            # TODO: aqui debemos colocar las preguntas que se puedan sacar del sampling corpus

        return update_button_clicked

    update_simulator_btn.on_click(get_update_button_clicked())
    update_simulator_box = widgets.VBox(
        [update_simulator_html, update_simulator_corpus_list, update_simulator_btn],
        layout=Layout(width="100%", align_items="center"),
    )
    # /Update simulator #

    # Update simulator embedded #
    update_simulator_embedded_sampling_corpus_text = widgets.Text(
        value="True",
        description="Ejecutar sampling corpus",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_corpus_list = widgets.Text(
        value="1",
        description="Corpus Id's (csv)",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_metadata_token = widgets.Text(
        value="->",
        description="Metadata token separator",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_person = widgets.Text(
        description="Su Id",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_mammut_logger = widgets.Text(
        description="Id del Mammut que reporta",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_compilation_room = widgets.Text(
        description="El ID de su room de reporte",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_simulator_embedded_regional_settings = widgets.Text(
        value="en",
        description="Idioma de reporte",
        style={"description_width": "50%"},
        layout=Layout(width="100%"),
    )
    update_embedded_button = widgets.Button(
        description="Actualizar Mammut Embedded", button_style="primary"
    )

    def __refresh_output(b):
        thoughts_list_template = get_template("thoughts_list.html")
        thoughts = simulator_obj.mammut_api.get_thoughts_since(
            int(update_simulator_embedded_compilation_room.value),
            __refresh_output.last_date,
        )
        thoughts = [thought.text for thought in thoughts]
        context_dict = {"thoughts_list": thoughts}
        thoughts_list_html = thoughts_list_template.render(context_dict)
        if len(thoughts) > 0:
            __refresh_output.last_date = (
                mammut.get_str_format_timezone_aware_datetime_now()
            )
            update_simulator_html.value = (
                update_simulator_html.value + "<br>" + thoughts_list_html
            )

    __refresh_output.last_date = mammut.get_str_format_timezone_aware_datetime_now()

    def update_embedded_response_eval(r):
        if r is 200:
            update_simulator_html.value = (
                "Iniciando proceso exitoso de compilacion...<br><br>"
            )
        elif r is 428:
            update_simulator_html.value = "El room de reporte no es un room entre el usuario y el bot de reporte.<br><br>"
        elif r is 503:
            update_simulator_html.value = "Se ha detectado otra compilacion en curso. Intente nuevamente en unos minutos...<br><br>"
        else:
            update_simulator_html.value = (
                "Error al comunicarse con Mammut API...<br><br>"
            )

        update_embedded_button.description = "Refresh output"
        update_embedded_button.on_click(update_embedded_button_clicked, True)
        update_embedded_button.on_click(__refresh_output)
        return

    def update_embedded_button_clicked(b):
        compile_payload = dict()
        compile_payload["packageId"] = simulator_obj.package.main_id
        compile_payload["executeSamplingCorpus"] = (
            update_simulator_embedded_sampling_corpus_text.value == "True"
        )
        compile_payload["corpusIds"] = [
            int(x) for x in update_simulator_embedded_corpus_list.value.split(",")
        ]
        compile_payload[
            "slideMetadataTokenSeparator"
        ] = update_simulator_embedded_metadata_token.value
        compile_payload["mammutId"] = simulator_obj.current_user.id
        compile_payload["userId"] = int(update_simulator_embedded_person.value)
        compile_payload["roomId"] = int(
            update_simulator_embedded_compilation_room.value
        )
        compile_payload[
            "regionalSettings"
        ] = update_simulator_embedded_regional_settings.value
        compile_payload["mammutLoggerId"] = int(
            update_simulator_embedded_mammut_logger.value
        )
        compile_payload_json = json.dumps(compile_payload)
        post_compilation_response = simulator_obj.mammut_api.post_compilation(
            compile_payload_json
        )
        __last_date = mammut.get_str_format_timezone_aware_datetime_now()
        update_embedded_response_eval(post_compilation_response)

    update_embedded_button.on_click(update_embedded_button_clicked)
    update_simulator_embedded_box = widgets.VBox(
        [
            update_simulator_embedded_sampling_corpus_text,
            update_simulator_embedded_corpus_list,
            update_simulator_embedded_metadata_token,
            update_simulator_embedded_person,
            update_simulator_embedded_mammut_logger,
            update_simulator_embedded_compilation_room,
            update_simulator_embedded_regional_settings,
            update_simulator_html,
            update_embedded_button,
        ],
        layout=Layout(width="100%", align_items="center"),
    )
    # /Update simulator embedded #

    # Image retraining #
    initial_image_message = widgets.HTML(
        value="<b>No package selected</b>",
        placeholder="Message: ",
        description="Message: ",
    )
    retrain_images_box = widgets.VBox([initial_image_message])
    # /Image retraining #

    if simulator_obj.embedded:
        return tuple(
            [
                select_mammut_box,
                validate_package_box,
                update_simulator_embedded_box,
                retrain_images_box,
            ]
        )
    else:
        return tuple(
            [
                select_mammut_box,
                validate_package_box,
                update_simulator_box,
                retrain_images_box,
            ]
        )



from mammut.common.lexicon.dictionary import Dictionary
from mammut.common.lexicon.linguistic_standard import LinguisticStandard
from typing import Set, List
from collections import namedtuple
from mammut.common.synthetic.synthetic_corpus_generator.token_generator import TokenGenerator

BoundMorpheme = namedtuple(
    "BoundMorpheme", ["free", "bound", "mixed", "deleted", "added"]
)


class MorphemesGenerator(TokenGenerator):
    """General morphemes generator. Encapsulates bounded morphemes
    generator heuristics for each language.

    TODO: Only 'es' regional settings is supported.
    This module code might be completely changed with new
    morphological paradigms syntax.
    """

    ES_GENRE_VOCALS: Set = {"a", "o"}

    # Baseline models used in nouns. These should not be
    # considered when creating the list of bounded morphemes
    # since they don't have a free morpheme associated.
    NOUNS_BASELINE_MODELS: Set[str] = {
        "NONS-SIMPLE",
        "MS-SIMPLE",
        "NONP-SIMPLE",
        "MP-SIMPLE",
        "PRE_PART-ing",
    }

    # Baseline models used in adjectives. These should not be
    # considered when creating the list of bounded morphemes
    # since they don't have a free morpheme associated.
    ADJECTIVES_BASELINE_MODELS: Set[str] = {
        "ADJECTIVE_SING_MASC",
        "ADJECTIVE_PLU_MASC",
        "EXCLAMATIVE_SING",
        "EXCLAMATIVE_PLU",
        "INDETER_SING",
        "INDETER_PLU",
        "INTERROGATIVE_SING",
        "INTERROGATIVE_PLU",
        "NUMERAL_ADJ_SING",
        "NUMERAL_ADJ_PLU",
        "POSSESSIVE_ADJ_SING",
        "POSSESSIVE_ADJ_PLU",
        "RELATIVE_SING",
        "RELATIVE_PLU",
        "CUANTITATIVO_SING",
        "CUANTITATIVO_PLU",
        "CUYO_RELATIVE",
    }

    # Verbs models used for the retrieval of
    # verbs bounded morphemes
    VERBS_INCLUDED_MODELS: Set[str] = {"AMAR_3", "TEMER_4"}

    def __init__(self, standard: LinguisticStandard,
                 dictionary: Dictionary):
        self.standard = standard
        self.dictionary = dictionary

    # Overriding abstract method
    def get_tokens(
            self, pos: str = "", regional_settings: str = "es"
    ) -> List[BoundMorpheme]:
        """Return the List[BoundMorpheme] that results
        from the application of Standard morpheme paradigms
        to word models.
        """
        if pos == "NOUN":
            return self._generate_nouns(regional_settings)
        elif pos == "ADJ":
            return self._generate_adjectives(regional_settings)
        elif pos == "VERB":
            return self._generate_verbs(regional_settings)
        else:
            return []

    def _get_nouns_morpheme_paradigms(self, regional_settings: str):
        paradigms_producers = self.standard.morpheme_paradigms["NOUN"][
            regional_settings
        ]
        paradigms_producers = dict(
            filter(
                lambda paradigm: not paradigm[0] in self.NOUNS_BASELINE_MODELS,
                paradigms_producers.producers.items(),
            )
        )
        return paradigms_producers

    def _get_adjectives_morpheme_paradigms(self, regional_settings: str):
        # Remove baseline paradigms producers
        paradigms_producers = self.standard.morpheme_paradigms["ADJ"][
            regional_settings
        ]
        paradigms_producers = dict(
            filter(
                lambda paradigm: not paradigm[0]
                                     in self.ADJECTIVES_BASELINE_MODELS,
                paradigms_producers.producers.items(),
            )
        )
        return paradigms_producers

    def _get_verbs_morpheme_paradigms(self, regional_settings: str):
        paradigms_producers = self.standard.morpheme_paradigms["VERB"][
            regional_settings
        ]
        paradigms_producers = dict(
            filter(
                lambda paradigm: paradigm[0] in self.VERBS_INCLUDED_MODELS,
                paradigms_producers.producers.items(),
            )
        )
        return paradigms_producers

    def _generate_nouns(self, regional_settings: str):
        # Remove baseline paradigms producers
        paradigms_producers = self._get_nouns_morpheme_paradigms(regional_settings)
        ret = []
        for key, val in paradigms_producers.items():
            self._process_noun_paradigms(key, val.inflection_rules, ret)

        return ret

    def _generate_adjectives(self, regional_settings: str):
        paradigms_producers = self._get_adjectives_morpheme_paradigms(regional_settings)
        ret = []
        for key, val in paradigms_producers.items():
            self._process_adjective_paradigms(key, val.inflection_rules, ret)

        return ret

    def _generate_verbs(self, regional_settings: str):
        # Keep only relevant paradigms
        paradigms_producers = self._get_verbs_morpheme_paradigms(regional_settings)
        ret = []
        for key, val in paradigms_producers.items():
            self._process_verb_paradigms(key.split("_")[0], val.inflection_rules, ret)

        return ret

    def get_free_lemma_from_dictionary(self) -> List[BoundMorpheme]:
        """Returns the List[BoundMorphemes] that results from
        the application of Standard morpheme paradigms to the
        lexicon lemmas.
        """
        lemmas = self.dictionary.lemmas_entries
        lemmas = list(filter(
            lambda lemma: lemma.postags_entries[0].pos_tag in ["VERB", "NOUN", "ADJ"],
            lemmas
        ))
        lemmas_nouns = list(filter(
            lambda lemma: lemma.postags_entries[0].pos_tag == "NOUN" and
                not lemma.postags_entries[0].morpheme_paradigm in self.NOUNS_BASELINE_MODELS,
            lemmas
        ))
        lemmas_adjectives = list(filter(
            lambda lemma: lemma.postags_entries[0].pos_tag == "ADJ" and
                          not lemma.postags_entries[0].morpheme_paradigm in self.ADJECTIVES_BASELINE_MODELS,
            lemmas
        ))
        lemmas_verbs = list(filter(
            lambda lemma: lemma.postags_entries[0].pos_tag == "VERB" and
                          lemma.postags_entries[0].morpheme_paradigm in self.VERBS_INCLUDED_MODELS,
            lemmas
        ))
        nouns_paradigms_producers = self._get_nouns_morpheme_paradigms('es')
        nouns_ret = []
        for noun_lemma in lemmas_nouns:
            pp = nouns_paradigms_producers[noun_lemma.postags_entries[0].morpheme_paradigm]
            self._process_noun_paradigms(noun_lemma.lemma, pp.inflection_rules, nouns_ret)

        adjectives_paradigms_producers = self._get_adjectives_morpheme_paradigms('es')
        adjectives_ret = []
        for adjective_lemma in lemmas_adjectives:
            pp = adjectives_paradigms_producers[adjective_lemma.postags_entries[0].morpheme_paradigm]
            self._process_adjective_paradigms(adjective_lemma.lemma, pp.inflection_rules, adjectives_ret)

        verbs_paradigms_producers = self._get_verbs_morpheme_paradigms('es')
        verbs_ret = []
        for verb_lemma in lemmas_verbs:
            pp = verbs_paradigms_producers[verb_lemma.postags_entries[0].morpheme_paradigm]
            self._process_verb_paradigms(verb_lemma.lemma, pp.inflection_rules, verbs_ret)

        return nouns_ret + adjectives_ret + verbs_ret

    def _process_noun_paradigms(self, word, rules, ret_list):
        """Apply an heuristic for nouns, for identification
        of free/bound morphemes in the word forms.

        Note: used for 'es' regional settings

        It mutates the reference object ret_list.
        """
        word_model = word.lower()
        for rule in rules:
            morpheme = rule.apply(word_model)
            bound = ""
            free = ""
            mixed = False
            if (
                    word_model == morpheme.text
                    and word_model[len(word_model) - 1] in self.ES_GENRE_VOCALS
            ):
                free = word_model[0: len(word_model) - 1]
                bound = word_model[len(word_model) - 1]
            elif word_model == morpheme.text:
                free = morpheme.text
                mixed = True
            elif not morpheme.deleted:
                if (
                        word_model[len(word_model) - 1] in self.ES_GENRE_VOCALS
                        or word_model[len(word_model) - 1] == "e"
                ):
                    bound = "".join(
                        list(word_model[len(word_model) - 1])
                        + morpheme.added
                    )
                    free = word_model[0: len(word_model) - 1]
                else:
                    bound = "".join(morpheme.added)
                    free = word_model
            elif len(morpheme.deleted) == 1 and (
                    morpheme.deleted[0] == "z"
                    or morpheme.deleted[0] in self.ES_GENRE_VOCALS
            ):
                bound = "".join(morpheme.added)
                free = word_model[0: len(word_model) - 1]
            else:
                bound = "".join(morpheme.deleted + morpheme.added)
                free = word_model[
                       0: len(word_model) - len(morpheme.deleted)
                       ]
            ret_list.append(
                BoundMorpheme(
                    free, bound, mixed, morpheme.deleted, morpheme.added
                )
            )

    def _process_adjective_paradigms(self, word, rules, ret_list):
        """Apply an heuristic for adjectives, for identification
        of free/bound morphemes in the word forms.

        Note: used for 'es' regional settings

        It mutates the reference object ret_list.
        """
        word_model = word.lower()
        for rule in rules:
            morpheme = rule.apply(word_model)
            bound = ""
            free = ""
            mixed = False
            if (
                    word_model == morpheme.text
                    and word_model[len(word_model) - 1] in self.ES_GENRE_VOCALS
            ):
                free = word_model[0: len(word_model) - 1]
                bound = word_model[len(word_model) - 1]
            elif word_model == morpheme.text:
                free = morpheme.text
                mixed = True
            elif not morpheme.deleted:
                if (
                        word_model[len(word_model) - 1] in self.ES_GENRE_VOCALS
                        or word_model[len(word_model) - 1] == "u"
                        or word_model[len(word_model) - 1] == "i"
                ):
                    bound = "".join(
                        list(word_model[len(word_model) - 1])
                        + morpheme.added
                    )
                    free = word_model[0: len(word_model) - 1]
                else:
                    bound = "".join(morpheme.added)
                    free = word_model
            elif len(morpheme.deleted) == 1 and (
                    morpheme.deleted[0] == "z"
                    or morpheme.deleted[0] in self.ES_GENRE_VOCALS
                    or morpheme.deleted[0] == "e"
            ):
                bound = "".join(morpheme.added)
                free = word_model[0: len(word_model) - 1]
            else:
                bound = "".join(morpheme.deleted + morpheme.added)
                free = word_model[
                       0: len(word_model) - len(morpheme.deleted)
                       ]
            ret_list.append(
                BoundMorpheme(
                    free, bound, mixed, morpheme.deleted, morpheme.added
                )
            )

    def _process_verb_paradigms(self, word, rules, ret_list):
        """Apply an heuristic for verbs for identification
        of free/bound morphemes in the word forms.

        Note: used for 'es' regional settings

        It mutates the reference object ret_list.
        """
        word_model = word.lower()
        for rule in rules:
            morpheme = rule.apply(word_model)
            bound = ""
            free = ""
            mixed = False
            if not morpheme.deleted:
                bound = "".join(
                    list(word_model[len(word_model) - 2:])
                    + morpheme.added
                )
                free = word_model[0: len(word_model) - 2]
            else:
                if len(morpheme.deleted) == 2:
                    bound = "".join(morpheme.added)
                    free = word_model[
                           0: len(word_model) - len(morpheme.deleted)
                           ]
                elif len(morpheme.deleted) == 1:
                    bound = "".join(
                        list(word_model[len(word_model) - 2])
                        + morpheme.added
                    )
                    free = word_model[0: len(word_model) - 2]

                else:
                    bound = "".join(
                        list(word_model[len(word_model) - 2:])
                        + morpheme.added
                    )
                    free = word_model[0: len(word_model) - 2]
            ret_list.append(
                BoundMorpheme(
                    free, bound, mixed, morpheme.deleted, morpheme.added
                )
            )


from abc import ABC
from mammut.curriculum.report.utils import Component
from mammut.curriculum.core.mammut_session_context import MammutSessionContext


class CurriculumBaseMessage(ABC):
    """Generic message to report in the curriculum core components.
    """

    def __init__(
        self, component: Component, message: str, report_level,
    ):
        """
        Args:
            component: Which curriculum component reports the message.
            message: message text.
            mammut_session_context: context with the curriculum session data.
            report_level: message report level.
        """
        self.component: component = component
        self.level = report_level
        self.message = message
        self.package_id = None
        self.session_id = None
        self.mammut_id = None

    def set_context(self, mammut_session_context: MammutSessionContext):
        self.package_id = mammut_session_context.package_id
        self.session_id = mammut_session_context.session_id
        self.mammut_id = mammut_session_context.mammut_id

    def __str__(self):
        f_str = f"Session:{self.session_id} - Mammut:{self.mammut_id} - Package: {self.package_id} - {self.message}"
        return f_str


class CurriculumBaseMessageWithCourse(CurriculumBaseMessage):
    """Generic message to report in the curriculum components where
    course ID is relevant.
    """

    def __init__(
        self, component: Component, course_id, message: str, report_level,
    ):
        super(CurriculumBaseMessageWithCourse, self).__init__(
            component, message, report_level
        )
        self.course_id = course_id

    def __str__(self):
        f_str = f"Session:{self.session_id} - Mammut:{self.mammut_id} - Package:{self.package_id} - Course:{self.course_id} - {self.message}"
        return f_str


class CurriculumBaseMessageWithLesson(CurriculumBaseMessageWithCourse):
    """Generic message to report in the curriculum components where
    lesson ID is relevant.
    """

    def __init__(
        self,
        component: Component,
        course_id,
        lesson_id,
        message: str,
        report_level,
    ):
        super(CurriculumBaseMessageWithLesson, self).__init__(
            component, course_id, message, report_level
        )
        self.lesson_id = lesson_id

    def __str__(self):
        f_str = f"Session:{self.session_id} - Mammut:{self.mammut_id} - Package:{self.package_id} - Course:{self.course_id} - Lesson:{self.lesson_id} - {self.message}"
        return f_str



# coding=utf-8
import os
import configparser
from pkg_resources import resource_string

__all__ = [
    "agenda",
    "agenda_app",
    "corpus_compiler_service",
    "InterpreterMain_pb2",
    "Protos_pb2",
    "WorkingMemory_pb2",
]

APPS_CONFIGURATIONS = configparser.ConfigParser()
APPS_CONFIGURATIONS.read_string(
    resource_string("mammut.resources", "apps.ini").decode("utf-8")
)
APPS_CONFIGURATIONS.read(["data/apps.ini", "apps.ini"])

KAFKA_BROKER_STR = (
    APPS_CONFIGURATIONS["KAFKA"]["host"] + ":" + APPS_CONFIGURATIONS["KAFKA"]["port"]
)
COMPILER_SERVICE_INPUT_TOPIC = APPS_CONFIGURATIONS["KAFKA"][
    "compiler_service_input_topic"
]
COMPILER_SERVICE_OUTPUT_TOPIC = APPS_CONFIGURATIONS["KAFKA"][
    "compiler_service_output_topic"
]
COMPILER_OUTPUT_RETRIES = int(APPS_CONFIGURATIONS["KAFKA"]["compiler_output_retries"])
COMPILER_OUTPUT_TIMEOUT = int(APPS_CONFIGURATIONS["KAFKA"]["compiler_output_timeout"])
COMPILER_OUTPUT_MAX_RECORDS = int(
    APPS_CONFIGURATIONS["KAFKA"]["compiler_output_max_records"]
)



# coding=utf-8
from typing import Union, Any, Dict, Optional

from elasticsearch import Elasticsearch, TransportError
import logging
import pandas as pd

from mammut import CONCORDANCE_QUERY_SIZE, ES_HOSTS
from mammut.common import (
    FEATURES_CONFIGURATIONS,
    index_definition_timeout,
    index_lemma_timeout,
    index_creation_timeout,
    global_request_timeout,
)
import os
import elasticsearch.helpers
import hashlib

from mammut.common.constants import DictionaryTable, ElasticsearchParams
from mammut.common.corpus.constants import CorpusTable

log = logging.getLogger(__name__)


class IndexInventory:
    INDEX_NAME = "datasets_inventory"
    INDEX_COLUMN = "index"
    DATASET_COLUMN = "dataset"
    HASH_COLUMN = "hash"

    def __init__(self):
        self.elasticsearch = Elasticsearch(ES_HOSTS)
        self._index_exist = False

    def _create_index_if_no_exist(self):
        if not self._index_exist:
            exist = self.elasticsearch.indices.exists(IndexInventory.INDEX_NAME)
            if not exist:
                self.elasticsearch.indices.create(IndexInventory.INDEX_NAME)
            self._index_exist = True

    def id_generator(self, index: str, dataset_name: str):
        return f"{index}_{dataset_name}"

    def set_index_data(
        self, index: str, dataset_name: str, data: Union[pd.DataFrame, Dict[str, Any]]
    ) -> Union[bool, int, str]:
        self._create_index_if_no_exist()
        data_hash = self.get_data_hash(data)
        response = self.elasticsearch.index(
            index=IndexInventory.INDEX_NAME,
            doc_type="doc",
            body={
                IndexInventory.INDEX_COLUMN: index,
                IndexInventory.DATASET_COLUMN: dataset_name,
                IndexInventory.HASH_COLUMN: data_hash,
            },
            id=self.id_generator(index, dataset_name,),
        )
        return response[ElasticsearchParams.RESULT_RESPONSE_KEY()]

    def get_data_hash(self, data):
        data_hash = None
        if isinstance(data, pd.DataFrame):
            data_hash = pd.util.hash_pandas_object(data)
            data_hash = hash(frozenset(data_hash.tolist()))
        elif isinstance(data, dict):
            data_hash = hash(frozenset(data.items()))
        else:
            data_hash = hash(data)
        return data_hash

    def get_index_data(self, index: str, dataset_name: str) -> Optional[str]:
        self._create_index_if_no_exist()
        data_hash = None
        exist = self.elasticsearch.exists(
            index=IndexInventory.INDEX_NAME,
            doc_type="doc",
            id=self.id_generator(index, dataset_name,),
        )
        if exist:
            response = self.elasticsearch.get(
                index=IndexInventory.INDEX_NAME,
                doc_type="doc",
                id=self.id_generator(index, dataset_name,),
            )
            if (
                response[ElasticsearchParams.FOUND_RESPONSE_KEY()]
                and ElasticsearchParams.SOURCE_RESPONSE_KEY() in response
                and IndexInventory.HASH_COLUMN
                in response[ElasticsearchParams.SOURCE_RESPONSE_KEY()]
            ):
                data_hash = response[ElasticsearchParams.SOURCE_RESPONSE_KEY()][
                    IndexInventory.HASH_COLUMN
                ]
        return data_hash

    def have_to_index_data(
        self, index: str, dataset_name: str, data: Union[pd.DataFrame, Dict[str, Any]]
    ) -> bool:
        expected_data_hash = self.get_data_hash(data)
        stored_data_hash = self.get_index_data(index, dataset_name)
        if stored_data_hash is not None and stored_data_hash == expected_data_hash:
            return False
        else:
            return True


CurrentIndexInventory = IndexInventory()


class CorpusIndexer:

    DEFAULT_COLUMN_TITTLES = [
        CorpusTable.ID_COLUMN_NAME(),
        CorpusTable.SUB_ID_COLUMN_NAME(),
        CorpusTable.SCENARIO_TYPE_COLUMN_NAME(),
        CorpusTable.EVENT_MESSAGE_COLUMN_NAME(),
        CorpusTable.HIDDEN_COLUMN_NAME(),
        CorpusTable.FIELD_COLUMN_NAME(),
        CorpusTable.LAMBDA_CONDITION_COLUMN_NAME(),
        CorpusTable.UI_EVENT_COLUMN_NAME(),
        CorpusTable.ACTION_COLUMN_NAME(),
        CorpusTable.SOURCE_COLUMN_NAME(),
        CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME(),
        CorpusTable.COMPLEXITY_COLUMN_NAME(),
    ]

    def __init__(self):
        self.elasticsearch = Elasticsearch(ES_HOSTS)

    def id_generator(
        self, spreadsheet_id, corpus_name, event_scenario_id, event_sub_id
    ):
        return spreadsheet_id + corpus_name + event_scenario_id + event_sub_id

    def index_event(self, index, spreadsheet_id, corpus_name, event):
        response = self.elasticsearch.index(
            index=index,
            doc_type="doc",
            body=event,
            id=self.id_generator(
                spreadsheet_id,
                corpus_name,
                event[CorpusTable.ID_COLUMN_NAME()],
                event[CorpusTable.SUB_ID_COLUMN_NAME()],
            ),
        )
        return response[ElasticsearchParams.RESULT_RESPONSE_KEY()]

    def delete_event(self, index, event_id):
        response = self.elasticsearch.delete(index=index, doc_type="doc", id=event_id)
        return response[ElasticsearchParams.FOUND_RESPONSE_KEY()]

    def delete_index(self, index):
        self.elasticsearch.indices.delete(index=index, ignore=[400, 404])

    def index_corpus(self, corpus_data_frame, spreadsheet_id, corpus_name, index):
        dataset_name = f"{spreadsheet_id}-{corpus_name}"
        have_to_index = CurrentIndexInventory.have_to_index_data(
            index, dataset_name, corpus_data_frame
        )
        if have_to_index:
            events = []
            for i in range(0, len(corpus_data_frame.index)):
                event = {}
                for column in CorpusIndexer.DEFAULT_COLUMN_TITTLES:
                    if corpus_data_frame.__contains__(column):
                        event[column] = corpus_data_frame[column][i]
                    else:
                        event[column] = ""
                event["corpus-name"] = corpus_name
                events.append(event)
            for event in events:
                self.index_event(index, spreadsheet_id, corpus_name, event)
            CurrentIndexInventory.set_index_data(index, dataset_name, corpus_data_frame)


class NGramIndexer:
    DEFAULT_REGIONAL_SETTINGS_KEY = "regional_settings"

    DEFAULT_PROVE_DATA_LOCATION_KEY = "prove_data_location"
    DEFAULT_DATA_LOCATION_KEY = "data_location"
    DEFAULT_DEFAULTNGRAM_KEY = "DEFAULT_NGRAM"
    DEFAULT_USERNGRAM_KEY = "USER_NGRAM"

    DEFAULT_COLUMN_TITTLES = ["ngram", "year", "match_count", "volume_count"]

    def __init__(self):
        self.elasticsearch = Elasticsearch(ES_HOSTS)

    def id_generator(self, ngram, year, path):
        return hashlib.md5((path + ngram + year).encode()).hexdigest()

    def index_ngram(self, index, year, ngram, path, ngram_doc):
        response = self.elasticsearch.index(
            index=index,
            doc_type="doc",
            body=ngram_doc,
            id=self.id_generator(ngram, year, path),
        )

        return response[ElasticsearchParams.RESULT_RESPONSE_KEY()]

    def delete_ngram(self, index, ngram_id):
        response = self.elasticsearch.delete(index=index, doc_type="doc", id=ngram_id)
        return response[ElasticsearchParams.FOUND_RESPONSE_KEY()]

    def delete_index(self, index):
        self.elasticsearch.indices.delete(index=index, ignore=[400, 404])

    def load_file(self, path, regional_settings):
        file = open(path)
        ngram_file = {}
        for line in file.readlines():
            ngram_elem = line.split("\t")
            ngram_file[self.id_generator(ngram_elem[0], ngram_elem[1], path)] = {
                NGramIndexer.DEFAULT_COLUMN_TITTLES[0]: ngram_elem[0],
                NGramIndexer.DEFAULT_COLUMN_TITTLES[1]: int(ngram_elem[1]),
                NGramIndexer.DEFAULT_COLUMN_TITTLES[2]: int(ngram_elem[2]),
                NGramIndexer.DEFAULT_COLUMN_TITTLES[3]: int(
                    ngram_elem[3].split("\n")[0]
                ),
                NGramIndexer.DEFAULT_REGIONAL_SETTINGS_KEY: regional_settings,
            }
        return ngram_file

    def index_ngram_file(self, file, index, prove_data=False):
        path = ""
        regional_settings = file.split("-")[1]
        DATA_LOCATION_KEY = ""

        if prove_data:
            DATA_LOCATION_KEY = NGramIndexer.DEFAULT_PROVE_DATA_LOCATION_KEY
        else:
            DATA_LOCATION_KEY = NGramIndexer.DEFAULT_DATA_LOCATION_KEY

        if (
            os.system(
                "cd "
                + os.path.join(
                    os.path.expanduser("~"),
                    FEATURES_CONFIGURATIONS[NGramIndexer.DEFAULT_DEFAULTNGRAM_KEY][
                        DATA_LOCATION_KEY
                    ],
                )
            )
            == 0
        ):
            os.chdir(os.path.expanduser("~"))
            path = (
                "./"
                + FEATURES_CONFIGURATIONS[NGramIndexer.DEFAULT_DEFAULTNGRAM_KEY][
                    DATA_LOCATION_KEY
                ]
                + "/"
                + file
            )
        else:
            os.chdir(os.path.expanduser("~"))
            path = (
                "./"
                + FEATURES_CONFIGURATIONS[NGramIndexer.DEFAULT_USERNGRAM_KEY][
                    DATA_LOCATION_KEY
                ]
                + "/"
                + file
            )

        ngram_file = self.load_file(path, regional_settings)
        dataset_name = f"ngram-{file}"
        have_to_index = CurrentIndexInventory.have_to_index_data(
            index, dataset_name, ngram_file
        )
        if have_to_index:
            for ngram in ngram_file.values():
                self.index_ngram(
                    index,
                    str(ngram[NGramIndexer.DEFAULT_COLUMN_TITTLES[1]]),
                    ngram[NGramIndexer.DEFAULT_COLUMN_TITTLES[0]],
                    path,
                    ngram,
                )
            CurrentIndexInventory.set_index_data(index, dataset_name, ngram_file)

    def file_generator(self, file, index, prove_data=False):
        path = ""
        regional_settings = file.split("-")[1]

        DATA_LOCATION_KEY = ""

        if prove_data:
            DATA_LOCATION_KEY = NGramIndexer.DEFAULT_PROVE_DATA_LOCATION_KEY
        else:
            DATA_LOCATION_KEY = NGramIndexer.DEFAULT_DATA_LOCATION_KEY
        user_path = os.path.expanduser("~")
        folder_path = os.path.join(
            user_path,
            FEATURES_CONFIGURATIONS[NGramIndexer.DEFAULT_DEFAULTNGRAM_KEY][
                DATA_LOCATION_KEY
            ],
        )
        if not os.path.exists(folder_path):
            folder_path = os.path.join(
                user_path,
                FEATURES_CONFIGURATIONS[NGramIndexer.DEFAULT_USERNGRAM_KEY][
                    DATA_LOCATION_KEY
                ],
            )
        path = folder_path + "/" + file
        file = open(path)
        line = None

        while True:
            line = file.readline()
            if not line:
                break
            ngram_elem = line.split("\t")
            ngram_doc = {
                "_op_type": "index",
                "_index": index,
                "_type": "doc",
                "_id": self.id_generator(ngram_elem[0], ngram_elem[1], path),
                "_source": {
                    NGramIndexer.DEFAULT_COLUMN_TITTLES[0]: ngram_elem[0],
                    NGramIndexer.DEFAULT_COLUMN_TITTLES[1]: int(ngram_elem[1]),
                    NGramIndexer.DEFAULT_COLUMN_TITTLES[2]: int(ngram_elem[2]),
                    NGramIndexer.DEFAULT_COLUMN_TITTLES[3]: int(
                        ngram_elem[3].split("\n")[0]
                    ),
                    NGramIndexer.DEFAULT_REGIONAL_SETTINGS_KEY: regional_settings,
                },
            }

            yield ngram_doc

    def index_ngram_file_bulk(self, file, index, prove_data=False):
        itor = self.file_generator(file, index, prove_data=prove_data)
        return elasticsearch.helpers.bulk(self.elasticsearch, itor)

    def index_ngram_file_parallel_bulk(self, file, index, prove_data=False):
        itor = self.file_generator(file, index, prove_data=prove_data)
        p_itor = elasticsearch.helpers.parallel_bulk(self.elasticsearch, itor)
        for response in p_itor:
            pass


class DefinitionIndexer:
    CORPUS_NAME = "definitions"

    def __init__(self, package_name):
        self.package_name = package_name
        self.definition_index_name = (
            self.package_name + "-" + DefinitionIndexer.CORPUS_NAME
        )
        self.elasticsearch = Elasticsearch(
            ES_HOSTS, request_timeout=global_request_timeout
        )
        if not self.elasticsearch.indices.exists(self.package_name):
            self.elasticsearch.indices.create(
                self.package_name,
                request_timeout=global_request_timeout,
                timeout=index_creation_timeout,
            )
        if not self.elasticsearch.indices.exists(self.definition_index_name):
            self.elasticsearch.indices.create(
                self.definition_index_name,
                request_timeout=global_request_timeout,
                timeout=index_creation_timeout,
            )

    def _index_definition(self, definition, morpheme):
        doc_id = (
            morpheme
            + "-"
            + definition[0]
            + "-"
            + definition[4]
            + "-"
            + str(definition[8])
        )
        self.elasticsearch.index(
            index=self.definition_index_name,
            doc_type="doc",
            body={
                DictionaryTable.LEMMA_COLUMN_NAME(): definition[0],
                DictionaryTable.WORD_TYPE_COLUMN_NAME(): definition[1],
                DictionaryTable.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME(): definition[2],
                DictionaryTable.AFI_COLUMN_NAME(): definition[3],
                DictionaryTable.POS_PARADIGM_COLUMN_NAME(): definition[4],
                DictionaryTable.MORPH_PARADIGM_COLUMN_NAME(): definition[5],
                DictionaryTable.MISS_PARADIGM_COLUMN_NAME(): definition[6],
                DictionaryTable.FET_PARADIGM_COLUMN_NAME(): definition[7],
                DictionaryTable.DEFINITION_INDEX_COLUMN_NAME(): definition[8],
                DictionaryTable.FEATURES_COLUMN_NAME(): definition[9],
                DictionaryTable.SOURCE_COLUMN_NAME(): definition[10],
                DictionaryTable.TOPIC_COLUMN_NAME(): definition[11],
                DictionaryTable.DEFINITION_COLUMN_NAME(): definition[12],
                DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME(): definition[
                    13
                ],
                DictionaryTable.DATE_COLUMN_NAME(): definition[14],
                "morpheme": morpheme,
            },
            id=doc_id,
            timeout=index_definition_timeout,
        )

    def index_lemma(self, definition_to_index, morphemes):
        """
        Index an entry in the dictionary as a corpus entry.
        The indexed document here used has the same exact structure
        of the one used in CorpusIndexer.

        :param definition_to_index: index of the definition inside the
        list of definitions for the lemma
        :param morphemes: all the formal morphemes associated to the lemma
        """
        doc_body_id = str(definition_to_index[0]) + "-" + str(definition_to_index[4])
        doc_id = doc_body_id + "-" + str(definition_to_index[8])
        try:
            self.elasticsearch.index(
                index=self.package_name,
                doc_type="doc",
                body={
                    CorpusTable.ID_COLUMN_NAME(): doc_body_id,
                    CorpusTable.SUB_ID_COLUMN_NAME(): str(definition_to_index[8]),
                    CorpusTable.SCENARIO_TYPE_COLUMN_NAME(): definition_to_index[
                        4
                    ],  # DictionaryTable.POS_PARADIGM_COLUMN_NAME()
                    CorpusTable.EVENT_MESSAGE_COLUMN_NAME(): definition_to_index[
                        12
                    ],  # DictionaryTable.DEFINITION_COLUMN_NAME()
                    CorpusTable.HIDDEN_COLUMN_NAME(): "",
                    CorpusTable.FIELD_COLUMN_NAME(): definition_to_index[
                        2
                    ],  # DictionaryTable.LEMMA_COLUMN_NAME()
                    CorpusTable.LAMBDA_CONDITION_COLUMN_NAME(): "",
                    CorpusTable.UI_EVENT_COLUMN_NAME(): "",
                    CorpusTable.ACTION_COLUMN_NAME(): "",
                    CorpusTable.SOURCE_COLUMN_NAME(): definition_to_index[
                        10
                    ],  # DictionaryTable.SOURCE_COLUMN_NAME()
                    CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME(): definition_to_index[
                        13
                    ],  # DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME()
                    CorpusTable.COMPLEXITY_COLUMN_NAME(): "",
                },
                id=doc_id,
                timeout=index_lemma_timeout,
            )
            for morpheme in morphemes:
                self._index_definition(definition_to_index, morpheme.text)
        except Exception as ex:
            mess = "Indexing Error: {}".format(ex)
            log.error(mess, exc_info=ex)
            raise ex

    def index_dictionary(
        self, dictionary, standard, spreadsheet_id: str, sheet_title: str,
    ):
        dataset_name = f"{spreadsheet_id}-{sheet_title}"
        have_to_index = CurrentIndexInventory.have_to_index_data(
            self.package_name, dataset_name, dictionary
        )
        if have_to_index:
            for index, lemma in enumerate(
                dictionary[DictionaryTable.LEMMA_COLUMN_NAME()]
            ):
                data = [
                    dictionary[DictionaryTable.LEMMA_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.WORD_TYPE_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.REGIONAL_SETTINGS_LEMMA_COLUMN_NAME()][
                        index
                    ],
                    dictionary[DictionaryTable.AFI_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.POS_PARADIGM_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.MORPH_PARADIGM_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.MISS_PARADIGM_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.FET_PARADIGM_COLUMN_NAME()][index],
                    int(
                        dictionary[DictionaryTable.DEFINITION_INDEX_COLUMN_NAME()][
                            index
                        ]
                    ),
                    dictionary[DictionaryTable.FEATURES_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.SOURCE_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.TOPIC_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.DEFINITION_COLUMN_NAME()][index],
                    dictionary[
                        DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME()
                    ][index],
                    dictionary[DictionaryTable.DATE_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.TIMESTAMP_COLUMN_NAME()][index],
                ]
                morphemes = standard.get_morphemes(
                    dictionary[DictionaryTable.POS_PARADIGM_COLUMN_NAME()][index],
                    dictionary[DictionaryTable.MORPH_PARADIGM_COLUMN_NAME()][
                        index
                    ].split(" ")[0],
                    dictionary[DictionaryTable.LEMMA_COLUMN_NAME()][index],
                    dictionary[
                        DictionaryTable.REGIONAL_SETTINGS_DEFINITION_COLUMN_NAME()
                    ][index],
                )
                self.index_lemma(data, morphemes)
            CurrentIndexInventory.set_index_data(
                self.package_name, dataset_name, dictionary
            )

    def get_receipt(self, lemma, definition):
        res = self.elasticsearch.search(
            index=self.definition_index_name,
            body={
                "query": {
                    "bool": {
                        "must": [
                            {
                                "match": {
                                    DictionaryTable.DEFINITION_COLUMN_NAME(): definition
                                }
                            },
                            {"match": {DictionaryTable.LEMMA_COLUMN_NAME(): lemma}},
                        ]
                    }
                }
            },
        )
        if res["hits"]["total"] > 0:
            return (
                "definition: "
                + definition
                + "\ncorrectly saved in spreadsheet with id: "
                + self.package_name
                + "\nfor lemma: "
                + lemma
            )
        else:
            return "Definition not found!!!"

    def get_defined_lemmas(self, date: str):
        try:
            docs = self.elasticsearch.search(
                index=self.definition_index_name,
                doc_type="doc",
                body={
                    "query": {"match": {DictionaryTable.DATE_COLUMN_NAME(): date}},
                    "size": CONCORDANCE_QUERY_SIZE,
                },
            )
            s = set()
            l_hits = docs["hits"]["hits"]
            for lemma in l_hits:
                s.add(lemma["_source"]["lemma"])
            return list(s)
        except TransportError as error:
            if error.error != "index_not_found_exception":
                print(str(error))
            return []



from brat_widget import Document

from mammut.common.lexicon.mammut_brat_configuration import (
    MammutBratConfiguration,
)


class MammutBratDocument(Document):
    def __init__(self, tokenized_message, config: MammutBratConfiguration):
        self.tokenized_message = tokenized_message
        # TODO: El manejo de offsets que espera el BRAT WIDGET no es el mismo que tenemos
        # Debemos definir correctamente esto.
        Document.__init__(
            self,
            self.tokenized_message.original_text,
            [],  # self.tokenized_message.token_offsets,
            self.tokenized_message.sentence_offsets,
            config,
        )



import re

#TODO: Eliminar Corpus_Transformer
class Corpus_Transformer:

    def __init__(self, thought_tuples: [], filters: []):
        self.tuples = thought_tuples
        self.filters = filters

    def execute(self, has_tokenized_tokens=False):

        for thought in self.tuples:
            for filter in self.filters:
                if not has_tokenized_tokens:
                    new_offers = filter.transform(thought.offers)
                    thought.offers = new_offers
                    print('len de ofers',len(thought.offers))
                else:
                    new_tokenized_offers = filter.transform(thought.tokenized_offers, has_tokenized_tokens)
                    thought.tokenized_offers = new_tokenized_offers
                    print('len de tokenized ofers',len(thought.tokenized_offers))

        return self.tuples

class Base_Corpus_Filter:
    TOKENIZED_SEPARATOR = 'toksep'

    def transform (self, messages:[], has_tokenized_tokens=False):

        if not has_tokenized_tokens:
            result = []
            for message in messages:
                result.extend(self.execute(message))
        else:
            result = []
            print('messages', messages)
            for message in messages:
                print('message individual', message)
                full_message = self.TOKENIZED_SEPARATOR.join(message)
                print('full message', full_message)
                result.extend(self.execute(full_message, has_tokenized_tokens))
                print('result', result)
                for i, val in enumerate(result):
                    if isinstance(val, str):
                        result[i] = val.split(self.TOKENIZED_SEPARATOR)
                print('final result', result)
                print('#####\n')
        return result

class Ignore_Accents_Filter (Base_Corpus_Filter):

    def __init__(self):
        self.accent_regex = r"[áéíóú]"
        self.accent_mapping = {'á': 'a', 'é': 'e', 'í': 'i', 'ó': 'o', 'ú': 'u'}

    def execute (self, orig_message:str, has_tokenized_tokens=None):

        result = []
        result.append(orig_message)

        punctuation_indices = [m.start(0) for m in re.finditer(self.accent_regex, orig_message)]

        for index in punctuation_indices:
            new_messages = []

            for message in result:
                #creates a new message changing the accent
                new_messages.append(message[:index] + self.accent_mapping[message[index]] + message[index + 1:])

            result.extend(new_messages)

        return result

class Ignore_Punctuation_Filter (Base_Corpus_Filter):

        def __init__(self):
            self.punctuation_regex = r"[!#$%&\'()*+,-./:;<=>?@^_`~¿¡]"

        def execute(self, orig_message: str, has_tokenized_tokens=None):

            result = []
            result.append(orig_message)

            punctuation_indices = [m.start(0) for m in re.finditer(self.punctuation_regex, orig_message)]
            orig_length = len(orig_message)

            for index in punctuation_indices:
                new_messages = []

                for message in result:
                    len_diff = orig_length - len(message)
                    # creates a new message without the given punctuation
                    new_messages.append(message[:index - len_diff] + message[index + 1 - len_diff:])

                result.extend(new_messages)

            return result

class Ignore_Casing_Filter (Base_Corpus_Filter):

    def execute(self, orig_message: str, has_tokenized_tokens=None):

        result = []
        result.append(orig_message)

        for index, char in enumerate(orig_message):

            if char.isalpha() is False:
                continue

            new_messages = []

            for message in result:
                # creates a new message swapping that character casing
                new_messages.append(message[:index] + char.swapcase() + message[index + 1:])

            result.extend(new_messages)

        return result

class Ignore_First_Letter_Casing_Filter (Base_Corpus_Filter):

    def execute(self, orig_message: str, has_tokenized_tokens=None):

        result = []
        result.append(orig_message)

        letter_index = 0

        if has_tokenized_tokens is None:
            #splits by whitespace
            splited_message = orig_message.split()
        else:
            splited_message = orig_message.split(self.TOKENIZED_SEPARATOR)

        for word in splited_message:

            new_messages = []

            for message in result:
                # creates a new message swapping the first character casing
                if word != '':
                    new_messages.append(message[:letter_index] + word[0].swapcase() + message[letter_index + 1:])
                else:
                    new_messages.append(message[:letter_index])

            if has_tokenized_tokens is None:
                letter_index += len(word) + 1
            else:
                letter_index += len(word) + len(self.TOKENIZED_SEPARATOR)
            result.extend(new_messages)

        return result




from mammut.common.basic_tokenizer import BasicTokenizer
from mammut.common.corpus.constants import CorpusTable
from mammut.common.corpus.corpus_base import (
    CorpusEventEntry,
    CorpusScenarioEntry,
    Corpus,
)
from mammut.common.lexicon.dictionary import Dictionary


class IssueCorpusEventEntry(CorpusEventEntry):
    def __init__(
        self,
        sub_id: int,
        str_message: str,
        str_source: str,
        str_issue_type: str,
        str_date: str,
        str_regional_settings: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        scenario_entry,
    ):
        CorpusEventEntry.__init__(
            self,
            sub_id,
            str_message,
            "",
            "",
            str_source,
            str_regional_settings,
            tokenizer,
            dictionary,
            scenario_entry,
        )

        self.issue_type = str_issue_type
        self.date = str_date


class IssueCorpusScenarioEntry(CorpusScenarioEntry):
    def __init__(
        self,
        str_id,
        str_type: str,
        scenerie_group_by_object,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        language_settings: str,
        corpus,
    ):
        CorpusScenarioEntry.__init__(
            self,
            str_id,
            str_type,
            scenerie_group_by_object,
            tokenizer,
            dictionary,
            language_settings,
            corpus,
        )

    def get_corpus_event_entry_from_event_row(self, index, e):
        event_mesage = ""
        if (
            CorpusTable.EVENT_MESSAGE_COLUMN_NAME() in e[1]
            and e[1][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()]
        ):
            event_mesage = e[1][CorpusTable.EVENT_MESSAGE_COLUMN_NAME()]

        return IssueCorpusEventEntry(
            index,
            event_mesage,
            e[1][CorpusTable.SOURCE_COLUMN_NAME()]
            if CorpusTable.SOURCE_COLUMN_NAME() in e[1]
            else "",
            e[1][CorpusTable.ISSUE_TYPE_COLUMN_NAME()],
            e[1][CorpusTable.DATE_COLUMN_NAME()],
            e[1][CorpusTable.REGIONAL_SETTINGS_COLUMN_NAME()],
            self.tokenizer,
            self.dictionary,
            self,
        )

    def validate_corpus_event_entry(
        self, index, cee, last_event_source, last_event_is_from_mammut
    ):
        if not cee.valid and not self.contains_invalid_events:
            self.valid = False
            self.contains_invalid_events = True
            self.errors.append("Contains invalid events")
            self.errors.extend(cee.errors)


class IssueCorpus(Corpus):
    def __init__(
        self,
        spreadsheet_id: str,
        id,
        corpus_sheet_title: str,
        corpus_extension_sheet_title: str,
        corpus_variables_sheet_title: str,
        tokenizer: BasicTokenizer,
        dictionary: Dictionary,
        regional_settings: str,
        annotable: bool,
    ):
        self.corpus_variables_sheet_title = corpus_variables_sheet_title
        Corpus.__init__(
            self,
            spreadsheet_id,
            id,
            corpus_sheet_title,
            corpus_extension_sheet_title,
            tokenizer,
            dictionary,
            regional_settings,
            annotable,
        )

    def get_corpus_scenario_entry_from_row(self, g):
        return IssueCorpusScenarioEntry(
            g[0][0],
            "Dialogue",
            g[1],
            self.tokenizer,
            self.dictionary,
            self.regional_settings,
            self,
        )



"""Message reporting package to be used across the curriculum system.

The package will export functions to simply start using the report functionality
right away, by just importing the package. For client classes, is as simple as:

    import mammut.curriculum.report as report
    report.send_message(Message("Hello"))


"""
from .reporters.client_reporter import ClientReporter
from .reporters.remote_client_reporter import RemoteClientReporter
from .reporters.ray_reporter_registry import RayReportersRegistry
from .reporters.console_reporter import ConsoleReporter
from .messages.curriculum_messages import *
from mammut.curriculum.core.mammut_session_context import MammutSessionContext
from mammut.config import settings
from typing import Optional
import logging

log = logging.getLogger(__name__)

__all__ = [
    "send_message",
    "RayReportersRegistry",
]

# Initializes the main client reporter and the remote client reporter.
# TODO: this classes must initialize from configuration values.
#   right now they are taking default hardcoded values in the class definition.

_main_reporter: ClientReporter = ClientReporter(
    threshold=logging.getLevelName(settings.curriculum.report_level)
)
_remote_reporter: RemoteClientReporter = RemoteClientReporter(
    threshold=logging.getLevelName(settings.curriculum.report_level)
)

# enable remote report. If this flag is true, all messages will be sent to a common registry actor handle.
_enable_remote: bool = False
_mammut_session_context: Optional[MammutSessionContext] = None


def send_message(message: CurriculumBaseMessage):
    """Sends a message to all the configured reporters.

    If remote report is enabled, reporter_registry_actor_handle is expected.

    Args:
        message: Message to be sent.
    """
    global _enable_remote
    global _main_reporter
    global _remote_reporter
    global _mammut_session_context

    message.set_context(_mammut_session_context)

    if _enable_remote:
        if (
            not _mammut_session_context
            or not _mammut_session_context.main_report_actor_handle
        ):
            log.error(
                "Trying to report remote message, but no registry actor handle found report context."
            )
            return

        _remote_reporter.send_message(
            message,
            **{
                "main_report_handle": _mammut_session_context.main_report_actor_handle
            },
        )

    else:
        _main_reporter.send_message(message)


def set_mammut_session_context(mammut_session_context: MammutSessionContext):
    """Set the session context for the messages reported in the
    python process.

    The MammutSessionContext contains the remote reporting actor handle
    for remote classrooms produced messages.

    Args:
        mammut_session_context:  mammut session context with the
            remote report actor handle.
    """
    global _mammut_session_context
    _mammut_session_context = mammut_session_context


def enable_remote():
    """Enable remote reporting in the classroom python process.
    """
    global _enable_remote
    _enable_remote = True



# coding=utf-8
from brat_widget.widget import Visualizer, Annotator
from brat_widget.document import *
from brat_widget.configuration import *
import pandas as pd
from mammut.common.storage.storage_manager import StorageManager


class Restitution:
    SHEET_TITTLE = "restitutions"

    def __init__(self, package, spreadsheet_id):
        self.package = package
        self.spreadsheet_id = spreadsheet_id
        self.storage_manager = StorageManager()

    def add_restitution(
        self, corpus_name, scenario_id, event_id, extension_id, restitution
    ):  # definir tipos, corpusEvent y str
        restitution_id = (
            corpus_name
            + "-"
            + str(scenario_id)
            + "-"
            + str(event_id)
            + "-"
            + str(extension_id)
        )
        row = [restitution_id, restitution]
        current_id = self.storage_manager.append_row_to_sheet(
            self.spreadsheet_id, Restitution.SHEET_TITTLE, row
        )
        # de aqui saco el id y se le agrega para saber cual es la restitucion del evento
        df = self.package.annotations.restitution_data_frame
        self.package.annotations.restitution_data_frame = df.append(
            pd.Series(row, list(df.columns)), ignore_index=True
        )
        return current_id

    def delete_restitutions(self, event_annotation):
        df = self.package.annotations.restitution_data_frame
        rows_indices = df[df["id"] == event_annotation.get_id()].index.tolist()
        rows_to_delete = len(rows_indices)
        self.package.annotations.restitution_data_frame = self.package.annotations.restitution_data_frame.drop(
            rows_indices
        )
        self.package.annotations.restitution_data_frame = self.package.annotations.restitution_data_frame.reset_index(
            drop=True
        )
        while rows_to_delete > 0:
            self.delete_restitution_from_sheet(rows_indices[0])
            rows_to_delete = rows_to_delete - 1
        deleted_restitutions = len(rows_indices)
        return deleted_restitutions

    def delete_restitution_from_sheet(self, row):
        response = self.storage_manager.delete_rows_from_spreadsheet(
            self.spreadsheet_id, "restitutions", str(row + 1), str(row + 2)
        )
        # TODO poner dos com constante es por el rowque ocupan los titulos y la cuenta del sheet empieza en 1
        return response


class MammutEventAnnotation:

    SCENARIO_INDEX_ID = "scenery"
    EVENT_ENTRY_INDEX_ID = "event_entry"

    def __init__(self, package, event_info):
        self.package = package
        self.standard = package.standard
        self.spreadsheet_id = package.spreadsheet_id
        self.dimensions = package.standard.dimensiones_descriptor_list
        self.dimensions_index = {k: i for i, k in enumerate(self.dimensions.keys())}
        self.event = event_info
        self.is_annotated_dimention = {}
        self.storage_manager = StorageManager()
        for i in self.dimensions:
            self.is_annotated_dimention[i] = False

        self.documents_by_dimension = {}
        self.configurations_by_dimension = {}
        self.range_dimension = {}

    def load_range(self, data_frame, dimension, annotations):
        matching_ids = data_frame[data_frame["id"] == self.get_id()].index.tolist()
        if len(matching_ids) > 0:

            if dimension not in annotations.sorted_ranges_by_dimension:
                annotations.sorted_ranges_by_dimension[dimension] = []
            range_order_object = {
                dimension: {
                    self.get_id(): {
                        "range": (matching_ids[0], matching_ids[-1]),
                        "order_index": len(
                            annotations.sorted_ranges_by_dimension[dimension]
                        ),
                    }
                }
            }
            if (
                range_order_object
                not in annotations.sorted_ranges_by_dimension[dimension]
            ):
                annotations.sorted_ranges_by_dimension[dimension].append(
                    range_order_object
                )
            if dimension not in annotations.ranges_by_dimension:
                annotations.ranges_by_dimension[dimension] = {}
            if self.get_id() not in annotations.ranges_by_dimension[dimension]:
                annotations.ranges_by_dimension[dimension][self.get_id()] = {
                    "range": (matching_ids[0], matching_ids[-1]),
                    "order_index": annotations.sorted_ranges_by_dimension[
                        dimension
                    ].index(range_order_object),
                }
            return

        if (
            dimension not in annotations.ranges_by_dimension
        ):  # TODO esto es necesario??? en la linea enterior habia un print(get_id()) revisar
            annotations.ranges_by_dimension[dimension] = {}
        annotations.ranges_by_dimension[dimension][self.get_id()] = {
            "range": None,
            "order_index": None,
        }

    def load_configurations_and_documents(self, dimensions):
        for i in dimensions:
            self.initialize_configuration(i)
            self.initialize_document(i, self.event)

    def update_documents_after_restitution(self, restitution):
        self.event["original_text"] = restitution + " " + self.event["original_text"]
        self.load_configurations_and_documents(self.dimensions)
        for i in self.dimensions:
            self.is_annotated_dimention[i] = False

    def update_documents_after_delete_restitution(self, deleted_restitutions: int):
        restitutions = self.event["original_text"].split(" ")[:deleted_restitutions]
        self.event["original_text"] = " ".join(
            self.event["original_text"].split(" ")[deleted_restitutions:]
        )
        self.load_configurations_and_documents(self.dimensions)
        for i in self.dimensions:
            self.is_annotated_dimention[i] = False
        return restitutions

    def initialize_configuration(self, dimension):
        self.configurations_by_dimension[
            dimension
        ] = self.standard.get_annotation_dimension_class(dimension)
        self.configurations_by_dimension[dimension].initialize()

    def initialize_document(self, dimension, event):
        corpus = self.package.corpus_map.get_corpus_by_type(
            event["corpus"]["type"], event["corpus"]["name"]
        )
        # TODO: this update the original_text from tokenized_mesage so is painted in the anotator/ same for visualizer
        corpus.sceneries[event[MammutEventAnnotation.SCENARIO_INDEX_ID]].events[
            event[MammutEventAnnotation.EVENT_ENTRY_INDEX_ID]
        ].tokenized_messages[event["tokenized_message_index"]].original_text = event[
            "original_text"
        ]
        # TODO: se deben devol;ver los diferentses mesaeges cuando los hay
        self.documents_by_dimension[dimension] = (
            corpus.sceneries[event[MammutEventAnnotation.SCENARIO_INDEX_ID]]
            .events[event[MammutEventAnnotation.EVENT_ENTRY_INDEX_ID]]
            .get_annotation_documents(self.configurations_by_dimension[dimension])[0]
        )
        self.documents_by_dimension[dimension].initialize()

    def get_annotator(self, dimension, event):
        annotator = Annotator(
            value=self.documents_by_dimension[dimension],
            collection_configuration=self.configurations_by_dimension[dimension],
            general_configuration=GeneralConfiguration(),
        )
        return annotator

    def get_visualizer(self, dimension, event):
        visualizer = Visualizer(
            value=self.documents_by_dimension[dimension],
            collection_configuration=self.configurations_by_dimension[dimension],
        )
        return visualizer

    def get_annotations_elements(self, dimension):
        dimension_index = list(self.dimensions).index(dimension)
        if dimension_index != 0:
            required_dimension_annotation = list(self.dimensions)[:dimension_index][-1]
        else:
            required_dimension_annotation = None
        annotation_elements = []
        if required_dimension_annotation is None:
            annotation_elements.append(self.get_annotator(dimension, self.event))
        elif self.is_annotated_dimention[
            required_dimension_annotation
        ]:  # itera hasta la dimension que se indique
            for index, i in enumerate(self.dimensions):
                if self.is_annotated_dimention[i] and i != dimension:
                    annotation_elements.append(self.get_visualizer(i, self.event))
                else:
                    annotation_elements.append(self.get_annotator(i, self.event))
                    break
        else:  # itera hasta la ultima dimension valida
            for index, i in enumerate(self.dimensions):
                if self.is_annotated_dimention[i] and index != len(self.dimensions) - 1:
                    annotation_elements.append(self.get_visualizer(i, self.event))
                else:
                    annotation_elements.append(self.get_annotator(i, self.event))
                    break
        return annotation_elements

    def current_dimension(self, dimension):
        dimension_index = list(self.dimensions).index(dimension)
        if dimension_index != 0:
            required_dimension_annotation = list(self.dimensions)[:dimension_index][-1]
        else:
            required_dimension_annotation = None
        if required_dimension_annotation is None:
            return dimension
        elif self.is_annotated_dimention[required_dimension_annotation]:
            return dimension

        for i in self.is_annotated_dimention:
            if not self.is_annotated_dimention[i]:
                return i
        return None

    def set_rows_for_persist(self, ann_obj):
        annotation_rows = []
        for index, i in enumerate(ann_obj.get_textbounds()):
            annotation_row = self.get_row_from_text_bound_annotation(i)
            annotation_rows.append(annotation_row)
        sorted_events = self.get_dependent_sorted_events(ann_obj)
        for i in sorted_events:
            annotation_rows = annotation_rows + self.get_rows_from_event_annotation(
                i, ann_obj
            )
        for i in ann_obj.get_attributes():
            annotation_rows.append(self.get_row_from_attribute_annotation(i, ann_obj))
        return annotation_rows

    def get_row_from_text_bound_annotation(self, entity):
        annotation_row = list()
        annotation_row.append(self.get_id())
        annotation_row.append(entity.text)
        annotation_row.append(
            str(entity.spans[0][0]) + "," + str(entity.spans[0][1])
        )  # TODO:siempre es uno??
        self.package.annotations.mammut_id_brat_id_dict[
            self.get_id_from_brat_entity(
                (entity.spans[0][0], entity.spans[0][1]), entity.type
            )
        ] = entity.id
        annotation_row.append(entity.type)
        annotation_row.append("")
        annotation_row.append("")
        annotation_row.append("")
        annotation_row.append("")
        annotation_row.append("")
        annotation_row.append(MammutAnnotations.ANNOTATION_ENTITY_TYPE)
        return annotation_row

    def get_rows_from_event_annotation(self, event, ann_obj):
        event_rows = []
        annotation_row = [self.get_id(), "", "", ""]
        for arg in event.args:
            annotation_row.append(
                self.get_id_from_brat_entity(
                    ann_obj.get_ann_by_id(event.trigger).spans[0],
                    ann_obj.get_ann_by_id(event.trigger).type,
                )
            )
            if arg[1].startswith(
                "E"
            ):  # TODO:como se que es un evento??validar si el objeto tiene algo para validar?
                entity = ann_obj.get_ann_by_id(ann_obj.get_ann_by_id(arg[1]).trigger)
                annotation_row.append(
                    self.get_id_from_brat_entity(entity.spans[0], entity.type)
                )
            else:
                annotation_row.append(
                    self.get_id_from_brat_entity(
                        ann_obj.get_ann_by_id(arg[1]).spans[0],
                        ann_obj.get_ann_by_id(arg[1]).type,
                    )
                )
            annotation_row.append("")
            annotation_row.append("")
            annotation_row.append("")
            annotation_row.append(MammutAnnotations.ANNOTATION_EVENT_TYPE)
            event_rows.append(annotation_row)
            annotation_row = [self.get_id(), "", "", ""]
        return event_rows

    def get_row_from_attribute_annotation(self, attribute, ann_obj):
        annotation_row = [self.get_id(), "", "", "", "", ""]
        annotation_row.append(attribute.type)
        if isinstance(ann_obj.get_ann_by_id(attribute.target), EventAnnotation):
            target_id = ann_obj.get_ann_by_id(attribute.target).trigger
        else:
            target_id = attribute.target
        annotation_row.append(
            self.get_id_from_brat_entity(
                ann_obj.get_ann_by_id(target_id).spans[0],
                ann_obj.get_ann_by_id(target_id).type,
            )
        )
        annotation_row.append(attribute.value)
        annotation_row.append(MammutAnnotations.ANNOTATION_FEATURE_TYPE)
        return annotation_row

    def get_dependent_sorted_events(self, ann_obj):
        nested_events = list()
        simple_events = list()
        for event in ann_obj.get_events():
            is_nested = False
            for arg in event.args:
                if arg[1].startswith("E"):  # TODO:si es un evento
                    is_nested = True
                    break
            if is_nested:
                nested_events.append(event)
            else:
                simple_events.append(event)
        # TODO:puedo meter esto en el segundo for en lugar de volver a iterar sobre los nestedevents
        event_ids = [
            event.id for event in nested_events
        ]  # TODO:podria ser un diccionario?
        event_ids_dict = {event.id: event for event in nested_events}
        for event in nested_events:
            for arg in event.args:
                if arg[1] in event_ids:  # si el argumento es parte de los nested events
                    arg_index = event_ids.index(arg[1])
                    current_event_index = event_ids.index(event.id)
                    if (
                        arg_index > current_event_index
                    ):  # si el indice/posicoin del argumento es mayor que la del evento actual cambiamos la posicion
                        event_ids[arg_index], event_ids[current_event_index] = (
                            event_ids[current_event_index],
                            event_ids[arg_index],
                        )
        sorted_nested_events = [
            event_ids_dict[id_index] for id_index in event_ids
        ]  # event_ids en este punto esta ordenado
        return simple_events + sorted_nested_events

    def get_id(self):
        annotation_id = (
            self.event["corpus"]["name"]
            + "-"
            + str(self.event[MammutEventAnnotation.SCENARIO_INDEX_ID])
            + "-"
            + str(self.event[MammutEventAnnotation.EVENT_ENTRY_INDEX_ID])
            + "-"
            + str(self.event["tokenized_message_index"])
        )
        return annotation_id

    def get_id_from_brat_entity(self, span: tuple, pos: str):
        return self.get_id() + "-" + str(span[0]) + "-" + str(span[1]) + "-" + pos

    def add_annotation_to_sheet(self, dimension):
        delete_result, num_of_deleted_cells = self.delete_annotation_from_sheet(
            dimension
        )
        rows = self.set_rows_for_persist(self.documents_by_dimension[dimension].ann_obj)
        new_range = []
        for index, row in enumerate(rows):
            result = self.storage_manager.append_row_to_sheet(
                self.spreadsheet_id,
                self.package.annotations.default_annotation_dimensions[dimension],
                row,
            )
            if len(rows) == 1:
                new_range.append(
                    int(
                        result["updates"]["updatedRange"].split(
                            MammutAnnotations.DEFAULT_ANNOTATIONS_END_CELL
                        )[1]
                    )
                    - 2
                )
                new_range.append(
                    int(
                        result["updates"]["updatedRange"].split(
                            MammutAnnotations.DEFAULT_ANNOTATIONS_END_CELL
                        )[1]
                    )
                    - 2
                )
            elif index == 0:
                new_range.append(
                    int(
                        result["updates"]["updatedRange"].split(
                            MammutAnnotations.DEFAULT_ANNOTATIONS_END_CELL
                        )[1]
                    )
                    - 2
                )
            elif index == len(rows) - 1:
                new_range.append(
                    int(
                        result["updates"]["updatedRange"].split(
                            MammutAnnotations.DEFAULT_ANNOTATIONS_END_CELL
                        )[1]
                    )
                    - 2
                )
        self.package.annotations.update_ranges(
            (new_range[0], new_range[1]),
            dimension,
            self.get_id(),
            jump=num_of_deleted_cells,
        )
        self.is_annotated_dimention[dimension] = True

    def delete_annotation_from_sheet(self, dimension):
        num_of_deleted_cells = 0
        if (
            self.package.annotations.ranges_by_dimension[dimension][self.get_id()][
                "range"
            ]
            is not None
        ):
            num_of_deleted_cells = (
                self.package.annotations.ranges_by_dimension[dimension][self.get_id()][
                    "range"
                ][-1]
                - self.package.annotations.ranges_by_dimension[dimension][
                    self.get_id()
                ]["range"][0]
            )
            delete_response = self.storage_manager.delete_rows_from_spreadsheet(
                self.spreadsheet_id,
                self.package.annotations.default_annotation_dimensions[dimension],
                str(
                    self.package.annotations.ranges_by_dimension[dimension][
                        self.get_id()
                    ]["range"][0]
                    + 1
                ),
                str(
                    self.package.annotations.ranges_by_dimension[dimension][
                        self.get_id()
                    ]["range"][-1]
                    + 2
                ),
            )  # TODO poner dos com constante es por el rowque ocupan los titulos y la cuenta del sheet empieza en 1
        else:
            delete_response = None
        return delete_response, num_of_deleted_cells


class MammutAnnotations:
    DEFAULT_RESTITUTIONS_START_COLUMN = "A"
    DEFAULT_RESTITUTIONS_START_ROW = "1"
    DEFAULT_RESTITUTIONS_END_CELL = "B"
    DEFAULT_ANNOTATIONS_START_COLUMN = "A"
    DEFAULT_ANNOTATIONS_START_ROW = "1"
    DEFAULT_ANNOTATIONS_END_CELL = "J"
    DEFAULT_RESTITUTIONS_SHEET_NAME = "restitutions"
    ANNOTATION_ENTITY_TYPE = "entity"
    ANNOTATION_EVENT_TYPE = "event"
    ANNOTATION_FEATURE_TYPE = "feature"
    NO_VARIABLE_ID_PARAMS = 3
    storage_manager = StorageManager()

    def __init__(self, package):
        self.package = package
        self.spreadsheet_id = package.spreadsheet_id
        self.event_annotations = []
        self.events_by_corpus = {}
        self.annotation_event_by_corpus = {}
        # TODO: poner restitutions strin en constante
        self.restitution_data_frame = self.storage_manager.get_spreadsheet_as_dataframe(
            self.spreadsheet_id,
            self.DEFAULT_RESTITUTIONS_SHEET_NAME,
            self.DEFAULT_RESTITUTIONS_START_COLUMN,
            self.DEFAULT_RESTITUTIONS_START_ROW,
            self.DEFAULT_RESTITUTIONS_END_CELL,
        )
        self.ranges_by_dimension = {}
        self.sorted_ranges_by_dimension = {}
        self.annotation_data_frames = {}
        self.mammut_id_brat_id_dict = {}
        self.mammut_id_event_brat_id_dict = {}
        self.default_annotation_dimensions = {}
        for dimension in package.standard.dimensiones_descriptor_list:
            self.default_annotation_dimensions[dimension] = "a-" + dimension
            self.sorted_ranges_by_dimension[dimension] = []
        self.load_annotations_data_frames()
        self.load_events()
        self.load_annotation_events()
        self.load_restitutions()
        self.load_events_configurations()
        self.load_saved_annotations()
        self.sort_ranges_list()

    def update_ranges(self, new_range, dimension, id, jump=None, only_delete=False):
        add_new_range_object = False

        if jump is None:
            jump = new_range[1] - new_range[0] + 1
        else:
            jump = jump + 1

        changed_index = self.ranges_by_dimension[dimension][id]["order_index"]
        if changed_index is None:
            changed_index = len(self.sorted_ranges_by_dimension[dimension]) - 1
            add_new_range_object = True
        for i in self.sorted_ranges_by_dimension[dimension][changed_index + 1 :]:
            # TODO las dos lineas que vienen modifican dos cosas diferentess????
            self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                "order_index"
            ] = (
                self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                    "order_index"
                ]
                - 1
            )
            i[dimension][list(i[dimension].keys())[0]]["order_index"] = (
                i[dimension][list(i[dimension].keys())[0]]["order_index"] - 1
            )
            update_range = []
            if (
                self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                    "range"
                ][0]
                - jump
                <= 0
            ):
                jump = self.ranges_by_dimension[dimension][
                    list(i[dimension].keys())[0]
                ]["range"][0]
            update_range.append(
                self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                    "range"
                ][0]
                - jump
            )
            update_range.append(
                self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                    "range"
                ][1]
                - jump
            )
            self.ranges_by_dimension[dimension][list(i[dimension].keys())[0]][
                "range"
            ] = tuple(update_range)
            i[dimension][list(i[dimension].keys())[0]]["range"] = tuple(update_range)
        if not add_new_range_object:
            range_object = self.sorted_ranges_by_dimension[dimension].pop(changed_index)
            range_object[dimension][id]["order_index"] = len(
                self.sorted_ranges_by_dimension[dimension]
            )
            range_object[dimension][id]["range"] = new_range
        else:
            range_object = {
                dimension: {
                    id: {
                        "range": new_range,
                        "order_index": len(self.sorted_ranges_by_dimension[dimension]),
                    }
                }
            }
        if only_delete:
            self.ranges_by_dimension[dimension][id]["range"] = None
            self.ranges_by_dimension[dimension][id]["order_index"] = None
        else:
            self.sorted_ranges_by_dimension[dimension].append(range_object)
            self.ranges_by_dimension[dimension][id]["range"] = new_range
            self.ranges_by_dimension[dimension][id]["order_index"] = (
                len(self.sorted_ranges_by_dimension[dimension]) - 1
            )

    # TODO: revisar si es necesario el arreglo sorted_blah_vlah, no se podra ordenar de manera mas sencilla el diccionario????
    def sort_ranges_list(self):  # TODO: mejorar el tiempo de ejecucion de este metodo
        bigest_index = 0
        aux_range = (0, 0)
        aux_list = []
        for dimension in self.default_annotation_dimensions:
            while len(self.sorted_ranges_by_dimension[dimension]) > 0:

                for index, j in enumerate(self.sorted_ranges_by_dimension[dimension]):
                    if (
                        j[dimension][list(j[dimension].keys())[0]]["range"][0]
                        >= aux_range[1]
                    ):
                        aux_range = j[dimension][list(j[dimension].keys())[0]]["range"]
                        bigest_index = index
                aux_list.append(
                    self.sorted_ranges_by_dimension[dimension].pop(bigest_index)
                )
                bigest_index = 0
                aux_range = (0, 0)

                # TODO:en la linea superior, se reduce el size del array y se guarda el valor de indice del mayor, y cuando termina busca un indice que ya no existe
            self.sorted_ranges_by_dimension[dimension] = aux_list[
                ::-1
            ]  # reversed aux list
            for order_index, elem in enumerate(
                self.sorted_ranges_by_dimension[dimension]
            ):
                for id in elem[dimension]:
                    elem[dimension][id]["order_index"] = order_index
                    self.ranges_by_dimension[dimension][id]["order_index"] = order_index
            aux_list = []

    def load_annotations_data_frames(self):
        for i in self.default_annotation_dimensions:
            self.annotation_data_frames[
                i
            ] = self.storage_manager.get_spreadsheet_as_dataframe(
                self.spreadsheet_id,
                self.default_annotation_dimensions[i],
                self.DEFAULT_ANNOTATIONS_START_COLUMN,
                self.DEFAULT_ANNOTATIONS_START_ROW,
                self.DEFAULT_ANNOTATIONS_END_CELL,
            )

    def load_events(self):
        for corpus in self.package.corpus_map.get_all_corpus():
            if corpus.annotable:
                self.events_by_corpus[corpus.sheet_title] = []
                event_message_info = {}
                for scenario in corpus.sceneries:
                    for event_entry_index, event_entry in enumerate(
                        corpus.sceneries[scenario].events
                    ):
                        for tokenized_message_index, tokenized_message in enumerate(
                            event_entry.tokenized_messages
                        ):
                            event_message_info["corpus"] = {
                                "type": type(corpus),
                                "name": corpus.sheet_title,
                            }  # TODO mejorar estructura de este "event", deberia definirse asi y aqui?
                            event_message_info[
                                MammutEventAnnotation.SCENARIO_INDEX_ID
                            ] = scenario
                            event_message_info[
                                MammutEventAnnotation.EVENT_ENTRY_INDEX_ID
                            ] = event_entry_index
                            event_message_info[
                                "tokenized_message_index"
                            ] = tokenized_message_index
                            event_message_info[
                                "original_text"
                            ] = tokenized_message.original_text
                            event_message_info[
                                "regnional_settings"
                            ] = event_entry.str_regional_settings
                            self.events_by_corpus[corpus.sheet_title].append(
                                event_message_info
                            )
                            event_message_info = {}

    def load_annotation_events(
        self,
    ):  # TODO: esto podria hacerse de una sola vez en metodo load_events
        for corpus_name in self.events_by_corpus:
            self.annotation_event_by_corpus[corpus_name] = []
            for event in self.events_by_corpus[corpus_name]:
                event_annotation = MammutEventAnnotation(self.package, event)
                for dimension in self.default_annotation_dimensions:
                    event_annotation.load_range(
                        self.annotation_data_frames[dimension], dimension, self
                    )
                self.annotation_event_by_corpus[corpus_name].append(event_annotation)

    def load_restitutions(self):
        for i in self.restitution_data_frame.iterrows():
            restitution_id = i[1][0]
            restitution = i[1][1]
            (
                corpus,
                scenario,
                event_entry,
                tokenized_message_index,
            ) = self.parse_restitution_id(restitution_id)
            annotation_event = self.get_event_by_parsed_id(
                corpus, scenario, event_entry, tokenized_message_index
            )
            self.add_restitution_to_anotation_event(restitution, annotation_event)

    def load_events_configurations(self):
        for i in self.annotation_event_by_corpus:
            for j in self.annotation_event_by_corpus[i]:
                j.load_configurations_and_documents(self.default_annotation_dimensions)

    def load_saved_annotations(self):
        for dimension in self.annotation_data_frames:
            if (
                dimension in self.annotation_data_frames
                and self.annotation_data_frames[dimension] is not None
            ):
                for j in self.annotation_data_frames[dimension].iterrows():
                    annotation_id = j[1][0]
                    text = j[1][1]
                    span = j[1][2]
                    pos = j[1][3]
                    trigger = j[1][4]
                    argument = j[1][5]
                    attr_type = j[1][6]
                    target = j[1][7]
                    attribute = j[1][8]
                    annotation_type = j[1][9]
                    (
                        corpus,
                        scenario,
                        event_entry,
                        tokenized_message_index,
                    ) = self.parse_restitution_id(annotation_id)
                    annotation_event = self.get_event_by_parsed_id(
                        corpus, scenario, event_entry, tokenized_message_index
                    )
                    self.add_annotation_to_document_by_dimension(
                        text,
                        span,
                        pos,
                        annotation_event,
                        dimension,
                        trigger,
                        argument,
                        attr_type,
                        target,
                        attribute,
                        annotation_type,
                    )

    def add_annotation_to_document_by_dimension(
        self,
        text,
        span,
        pos,
        annotation_event,
        dimension,
        trigger,
        argument,
        attr_type,
        target,
        attribute,
        annotation_type,
    ):
        argument_id = None
        document = annotation_event.documents_by_dimension[dimension]
        if annotation_type == MammutAnnotations.ANNOTATION_ENTITY_TYPE:
            tuple_span = tuple([int(i) for i in span.split(",")])
            self.mammut_id_brat_id_dict[
                annotation_event.get_id_from_brat_entity(tuple_span, pos)
            ] = document.ann_obj.get_new_id("T")
            annotation = TextBoundAnnotationWithText(
                [tuple_span],
                self.mammut_id_brat_id_dict[
                    annotation_event.get_id_from_brat_entity(tuple_span, pos)
                ],
                pos,
                text,
            )
            document.ann_obj.add_annotation(annotation)
        elif annotation_type == MammutAnnotations.ANNOTATION_EVENT_TYPE:
            brat_id = document.ann_obj.get_new_id("E")
            entity = document.ann_obj.get_ann_by_id(
                self.mammut_id_brat_id_dict[argument]
            )
            type = entity.type
            annotation = {}
            event_exists = False
            # TODO: no seria mejor crear las listas y hacer el inititlaize en vez de crear la intancia yo
            for i in document.ann_obj.get_events():
                if i.trigger == self.mammut_id_brat_id_dict[trigger]:
                    annotation = i
                    event_exists = True

            if not event_exists:
                self.mammut_id_event_brat_id_dict[trigger] = brat_id
                if argument in self.mammut_id_event_brat_id_dict:
                    argument_id = self.mammut_id_event_brat_id_dict[argument]
                else:
                    argument_id = self.mammut_id_brat_id_dict[argument]
                annotation = EventAnnotation(
                    self.mammut_id_brat_id_dict[trigger], [], brat_id, type, "\t"
                )
                annotation.add_argument(entity.type, argument_id)
                document.ann_obj.add_annotation(annotation)
            else:
                if argument in self.mammut_id_event_brat_id_dict:
                    argument_id = self.mammut_id_event_brat_id_dict[argument]
                else:
                    argument_id = self.mammut_id_brat_id_dict[argument]
                annotation.add_argument(entity.type, argument_id)
        elif annotation_type == MammutAnnotations.ANNOTATION_FEATURE_TYPE:
            brat_id = document.ann_obj.get_new_id("A")
            tail = ""
            if target in self.mammut_id_event_brat_id_dict.keys():
                triger = self.mammut_id_event_brat_id_dict[target]
            else:
                triger = self.mammut_id_brat_id_dict[target]
            annotation = AttributeAnnotation(
                triger, brat_id, attr_type, tail, attribute
            )
            document.ann_obj.add_annotation(annotation)
        annotation_event.is_annotated_dimention[dimension] = True

    def add_restitution_to_anotation_event(
        self, restitution: str, annotation_event: MammutEventAnnotation
    ):
        if annotation_event is None:
            return
        annotation_event.event["original_text"] = (
            restitution + " " + annotation_event.event["original_text"]
        )

    def get_event_by_parsed_id(
        self,
        corpus: str,
        scenario_index: int,
        event_entry_index: int,
        tokenized_message_index: int,
    ):
        for i in self.annotation_event_by_corpus[corpus]:
            if (
                i.event[MammutEventAnnotation.SCENARIO_INDEX_ID] == scenario_index
                and i.event[MammutEventAnnotation.EVENT_ENTRY_INDEX_ID]
                == event_entry_index
                and i.event["tokenized_message_index"] == tokenized_message_index
            ):
                return i

    def parse_restitution_id(self, restitution_id):
        words = restitution_id.split("-")
        corpus = "-".join(words[: len(words) - MammutAnnotations.NO_VARIABLE_ID_PARAMS])
        tokenized_message_index = int(words[-1])
        event_entry = int(words[-2])
        scenario = int(words[-3])
        return corpus, scenario, event_entry, tokenized_message_index

    def get_event_messages(self, corpus):
        for i in self.events_by_corpus:
            if i == corpus.sheet_title:
                return self.events_by_corpus[i]

    def get_annotation_event_messages(self, corpus):
        for i in self.annotation_event_by_corpus:
            if i == corpus.sheet_title:
                return self.annotation_event_by_corpus[i]

    def delete_all_annotations(self, event_annotation):
        for dimension in self.default_annotation_dimensions:
            if (
                self.ranges_by_dimension[dimension][event_annotation.get_id()]["range"]
                is not None
            ):
                (
                    result,
                    num_of_deleted_cells,
                ) = event_annotation.delete_annotation_from_sheet(dimension)
                range = self.package.annotations.ranges_by_dimension[dimension][
                    event_annotation.get_id()
                ]["range"]
                jump = range[1] - range[0]
                self.update_ranges(
                    None, dimension, event_annotation.get_id(), jump, only_delete=True
                )



from mammut.curriculum.report.reporters.base_reporter import BaseReporter
from mammut.curriculum.report.reporters.console_reporter import ConsoleReporter
from mammut.curriculum.report.messages.message import CurriculumBaseMessage
from typing import List
import logging

log = logging.getLogger(__name__)


class ClientReporter(BaseReporter):
    """Reporter class to support client reporting calls.

    The client reporter can hold arbitrary custom reporters
    to replicate message report.

    This client can enable module logging to also publish messages
    in the conventional way.

    This reporter can enable caching. When caching is enabled, the messages
    will be saved in an ordered sequence.

    TODO: constructor parameters must be replaced by configuration values.
    """

    def __init__(
        self,
        threshold=logging.INFO,
        enable_module_logging: bool = False,
        enable_caching: bool = False,
    ):
        """
        Args:
            threshold: threshold level for this reporter.
            enable_caching (bool): enable the conventional module logger replication
                of messages
            enable_caching (bool): enable store of messages in this reporter state.
        """
        super(ClientReporter, self).__init__(threshold)
        self._reporters: List[BaseReporter] = []
        self._module_logging_enable: bool = enable_module_logging
        self._cache: List[
            CurriculumBaseMessage
        ] = [] if enable_caching else None
        self._add_reporter(
            ConsoleReporter(
                "client_console_reporter", threshold=self.threshold
            )
        )
        log.debug("Client reported configured")

    def _add_reporter(self, reporter: BaseReporter):
        if reporter not in self._reporters:
            self._reporters.append(reporter)

    def send_message(self, message: CurriculumBaseMessage, **kwargs):
        """Replicates the message report to all custom reporters.

        :Args:
            message: message to be reported.
        """
        if self._module_logging_enable:
            log.log(message.level, str(message))

        if self._cache:
            self._cache.append(message)

        for reporter in self._reporters:
            reporter.send_message(message)
